import numpy as np
import numba
import scipy.optimize as sopt
import json

sin = np.sin
cos = np.cos

class imib_fisix_3ord_class: 

    def __init__(self): 

        self.t_end = 10.000000 
        self.Dt = 0.0010000 
        self.decimation = 10.000000 
        self.itol = 1e-6 
        self.Dt_max = 0.001000 
        self.Dt_min = 0.001000 
        self.solvern = 5 
        self.imax = 100 
        self.N_x = 3
        self.N_y = 10 
        self.N_z = 4 
        self.N_store = 10000 
        self.params_list = ['S_b', 'U_b', 'I_b', 'R_s', 'R_r', 'L_ls', 'L_lr', 'L_m', 'H_m', 'Omega_b', 'D', 'v_0', 'theta_0', 'X_l', 'omega_s', 'v_rd', 'v_rq'] 
        self.params_values_list  = [90000.0, 400.0, 129.9038105676658, 0.012937500000000001, 0.013500000000000002, 0.048375, 0.11025000000000001, 2.0812500000000003, 3.5, 314.1592653589793, 0.1, 1, 0.0, 0.05, 1.0, 0.0, 0.0] 
        self.inputs_ini_list = ['P_h', 'Q_h'] 
        self.inputs_ini_values_list  = [0.1, 0.0] 
        self.inputs_run_list = ['tau_m', 'B_c'] 
        self.inputs_run_values_list = [0.1, 0.0] 
        self.x_list = ['omega_e', 'psi_rd', 'psi_rq'] 
        self.y_list = ['psi_sd', 'psi_sq', 'i_sd', 'i_sq', 'i_rd', 'i_rq', 'P_h', 'Q_h', 'v_h', 'theta_h'] 
        self.xy_list = self.x_list + self.y_list 
        self.y_ini_list = ['psi_sd', 'psi_sq', 'i_sd', 'i_sq', 'i_rd', 'i_rq', 'tau_m', 'B_c', 'v_h', 'theta_h'] 
        self.xy_ini_list = self.x_list + self.y_ini_list 
        self.t = 0.0
        self.it = 0
        self.it_store = 0

        self.sopt_root_method='hybr'
        self.sopt_root_jac=True

        self.update() 


    def update(self): 

        self.N_steps = int(np.ceil(self.t_end/self.Dt)) 
        dt = [  
              ('t_end', np.float64),
              ('Dt', np.float64),
              ('decimation', np.float64),
              ('itol', np.float64),
              ('Dt_max', np.float64),
              ('Dt_min', np.float64),
              ('solvern', np.int64),
              ('imax', np.int64),
              ('N_steps', np.int64),
              ('N_store', np.int64),
              ('N_x', np.int64),
              ('N_y', np.int64),
              ('N_z', np.int64),
              ('t', np.float64),
              ('it', np.int64),
              ('it_store', np.int64),
              ('idx', np.int64),
              ('idy', np.int64),
              ('f', np.float64, (self.N_x,1)),
              ('f_ini', np.float64, (self.N_x,1)),
              ('x', np.float64, (self.N_x,1)),
              ('x_ini', np.float64, (self.N_x,1)),
              ('x_0', np.float64, (self.N_x,1)),
              ('g', np.float64, (self.N_y,1)),
              ('g_ini', np.float64, (self.N_y,1)),
              ('y', np.float64, (self.N_y,1)),
              ('y_ini', np.float64, (self.N_y,1)),
              ('y_0', np.float64, (self.N_y,1)),
              ('h', np.float64, (self.N_z,1)),
              ('Fx', np.float64, (self.N_x,self.N_x)),
              ('Fy', np.float64, (self.N_x,self.N_y)),
              ('Gx', np.float64, (self.N_y,self.N_x)),
              ('Gy', np.float64, (self.N_y,self.N_y)),
              ('Fx_ini', np.float64, (self.N_x,self.N_x)),
              ('Fy_ini', np.float64, (self.N_x,self.N_y)),
              ('Gx_ini', np.float64, (self.N_y,self.N_x)),
              ('Gy_ini', np.float64, (self.N_y,self.N_y)),
              ('T', np.float64, (self.N_store+1,1)),
              ('X', np.float64, (self.N_store+1,self.N_x)),
              ('Y', np.float64, (self.N_store+1,self.N_y)),
              ('Z', np.float64, (self.N_store+1,self.N_z)),
              ('iters', np.float64, (self.N_store+1,1)),
             ]

        values = [
                self.t_end,                          
                self.Dt,
                self.decimation,
                self.itol,
                self.Dt_max,
                self.Dt_min,
                self.solvern,
                self.imax,
                self.N_steps,
                self.N_store,
                self.N_x,
                self.N_y,
                self.N_z,
                self.t,
                self.it,
                self.it_store,
                0,                                     # idx
                0,                                     # idy
                np.zeros((self.N_x,1)),                # f
                np.zeros((self.N_x,1)),                # f_ini
                np.zeros((self.N_x,1)),                # x
                np.zeros((self.N_x,1)),                # x_ini
                np.zeros((self.N_x,1)),                # x_0
                np.zeros((self.N_y,1)),                # g
                np.zeros((self.N_y,1)),                # g_ini
                np.zeros((self.N_y,1)),                # y
                np.zeros((self.N_y,1)),                # y_ini
                np.zeros((self.N_y,1)),                # y_0
                np.zeros((self.N_z,1)),                # h
                np.zeros((self.N_x,self.N_x)),         # Fx   
                np.zeros((self.N_x,self.N_y)),         # Fy 
                np.zeros((self.N_y,self.N_x)),         # Gx 
                np.zeros((self.N_y,self.N_y)),         # Fy 
                np.zeros((self.N_x,self.N_x)),         # Fx_ini  
                np.zeros((self.N_x,self.N_y)),         # Fy_ini 
                np.zeros((self.N_y,self.N_x)),         # Gx_ini 
                np.zeros((self.N_y,self.N_y)),         # Fy_ini 
                np.zeros((self.N_store+1,1)),          # T
                np.zeros((self.N_store+1,self.N_x)),   # X
                np.zeros((self.N_store+1,self.N_y)),   # Y
                np.zeros((self.N_store+1,self.N_z)),   # Z
                np.zeros((self.N_store+1,1)),          # iters
                ]  

        dt += [(item,np.float64) for item in self.params_list]
        values += [item for item in self.params_values_list]

        for item_id,item_val in zip(self.inputs_ini_list,self.inputs_ini_values_list):
            if item_id in self.inputs_run_list: continue
            dt += [(item_id,np.float64)]
            values += [item_val]

        dt += [(item,np.float64) for item in self.inputs_run_list]
        values += [item for item in self.inputs_run_values_list]

        self.struct = np.rec.array([tuple(values)], dtype=np.dtype(dt))

    def load_params(self,data_input):

        if type(data_input) == str:
            json_file = data_input
            self.json_file = json_file
            self.json_data = open(json_file).read().replace("'",'"')
            data = json.loads(self.json_data)
        elif type(data_input) == dict:
            data = data_input

        self.data = data
        for item in self.data:
            self.struct[0][item] = self.data[item]
            self.params_values_list[self.params_list.index(item)] = self.data[item]



    def ini_problem(self,x):
        self.struct[0].x_ini[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        ini(self.struct,2)
        ini(self.struct,3)       
        fg = np.vstack((self.struct[0].f_ini,self.struct[0].g_ini))[:,0]
        return fg

    def dae_jacobian(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        run(0.0,self.struct,10)
        run(0.0,self.struct,11)       
        A_c = np.block([[self.struct[0].Fx,self.struct[0].Fy],
                        [self.struct[0].Gx,self.struct[0].Gy]])
        return A_c

    def ini_dae_jacobian(self,x):
        self.struct[0].x_ini[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        ini(self.struct,10)
        ini(self.struct,11)       
        A_c = np.block([[self.struct[0].Fx_ini,self.struct[0].Fy_ini],
                        [self.struct[0].Gx_ini,self.struct[0].Gy_ini]])
        return A_c

    def run_problem(self,x):
        t = self.struct[0].t
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        run(t,self.struct,2)
        run(t,self.struct,3)
        run(t,self.struct,10)
        run(t,self.struct,11)
        fg = np.vstack((self.struct[0].f,self.struct[0].g))[:,0]
        return fg

    def f_ode(self,x):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def f_odeint(self,x,t):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def f_ivp(self,t,x):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def Fx_ode(self,x):
        self.struct[0].x[:,0] = x
        run(self.struct,10)
        return self.struct[0].Fx

    def eval_A(self):
        
        Fx = self.struct[0].Fx
        Fy = self.struct[0].Fy
        Gx = self.struct[0].Gx
        Gy = self.struct[0].Gy
        
        A = Fx - Fy @ np.linalg.solve(Gy,Gx)
        
        self.A = A
        
        return A

    def eval_A_ini(self):
        
        Fx = self.struct[0].Fx_ini
        Fy = self.struct[0].Fy_ini
        Gx = self.struct[0].Gx_ini
        Gy = self.struct[0].Gy_ini
        
        A = Fx - Fy @ np.linalg.solve(Gy,Gx)
        
        
        return A

    def simulate(self,events,xy0=0):
        # simulation parameters
        self.struct[0].it = 0       # set time step to zero
        self.struct[0].it_store = 0 # set storage to zero
        self.struct[0].t = 0.0      # set time to zero
                    
        # initialization
        it_event = 0
        event = events[it_event]
        for item in event:
            self.struct[0][item] = event[item]
            
        
        ## compute initial conditions using x and y_ini 
        if xy0 == 0:
            xy0 = np.zeros(self.N_x+self.N_y)
        elif xy0 == 1:
            xy0 = np.ones(self.N_x+self.N_y)
        else:
            xy0 = xy0*np.ones(self.N_x+self.N_y)

        #xy = sopt.fsolve(self.ini_problem,xy0, jac=self.ini_dae_jacobian )
        if self.sopt_root_jac:
            sol = sopt.root(self.ini_problem, xy0, jac=self.ini_dae_jacobian, method=self.sopt_root_method)
        else:
            sol = sopt.root(self.ini_problem, xy0, method=self.sopt_root_method)

        self.initialization_ok = True
        if sol.success == False:
            print('initialization not found!')
            self.initialization_ok = False

            T = self.struct[0]['T'][:self.struct[0].it_store]
            X = self.struct[0]['X'][:self.struct[0].it_store,:]
            Y = self.struct[0]['Y'][:self.struct[0].it_store,:]
            Z = self.struct[0]['Z'][:self.struct[0].it_store,:]
            iters = self.struct[0]['iters'][:self.struct[0].it_store,:]

        if self.initialization_ok:
            xy = sol.x
            self.struct[0].x[:,0] = xy[0:self.N_x]
            self.struct[0].y[:,0] = xy[self.N_x:]

            ## y_ini to u_run
            for item in self.inputs_run_list:
                if item in self.y_ini_list:
                    self.struct[0][item] = self.struct[0].y_ini[self.y_ini_list.index(item)]

            ## u_ini to y_run
            for item in self.inputs_ini_list:
                if item in self.y_list:
                    self.struct[0].y[self.y_list.index(item)] = self.struct[0][item]
        
            ## solve selfem
            daesolver(self.struct)    # run until first event

            # simulation run
            for event in events[1:]:  
                for item in event:
                    self.struct[0][item] = event[item]
                daesolver(self.struct)    # run until next event
                
            
            # post process result    
            T = self.struct[0]['T'][:self.struct[0].it_store]
            X = self.struct[0]['X'][:self.struct[0].it_store,:]
            Y = self.struct[0]['Y'][:self.struct[0].it_store,:]
            Z = self.struct[0]['Z'][:self.struct[0].it_store,:]
            iters = self.struct[0]['iters'][:self.struct[0].it_store,:]
        
            self.T = T
            self.X = X
            self.Y = Y
            self.Z = Z
            self.iters = iters
            
        return T,X,Y,Z


@numba.njit(cache=True)
def run(t,struct,mode):

    # Parameters:
    S_b = struct[0].S_b
    U_b = struct[0].U_b
    I_b = struct[0].I_b
    R_s = struct[0].R_s
    R_r = struct[0].R_r
    L_ls = struct[0].L_ls
    L_lr = struct[0].L_lr
    L_m = struct[0].L_m
    H_m = struct[0].H_m
    Omega_b = struct[0].Omega_b
    D = struct[0].D
    v_0 = struct[0].v_0
    theta_0 = struct[0].theta_0
    X_l = struct[0].X_l
    omega_s = struct[0].omega_s
    v_rd = struct[0].v_rd
    v_rq = struct[0].v_rq
    
    # Inputs:
    tau_m = struct[0].tau_m
    B_c = struct[0].B_c
    
    # Dynamical states:
    omega_e = struct[0].x[0,0]
    psi_rd = struct[0].x[1,0]
    psi_rq = struct[0].x[2,0]
    
    # Algebraic states:
    psi_sd = struct[0].y[0,0]
    psi_sq = struct[0].y[1,0]
    i_sd = struct[0].y[2,0]
    i_sq = struct[0].y[3,0]
    i_rd = struct[0].y[4,0]
    i_rq = struct[0].y[5,0]
    P_h = struct[0].y[6,0]
    Q_h = struct[0].y[7,0]
    v_h = struct[0].y[8,0]
    theta_h = struct[0].y[9,0]
    
    # Differential equations:
    if mode == 2:


        struct[0].f[0,0] = (-D*omega_e + i_sd*psi_sq - i_sq*psi_sd + tau_m)/(2*H_m)
        struct[0].f[1,0] = Omega_b*(-R_r*i_rd - psi_rq*(-omega_e + omega_s) - v_rd)
        struct[0].f[2,0] = Omega_b*(-R_r*i_rq + psi_rd*(-omega_e + omega_s) - v_rq)
    
    # Algebraic equations:
    if mode == 3:


        struct[0].g[0,0] = Omega_b*(-R_s*i_sd - omega_s*psi_sq + v_h*sin(theta_h))
        struct[0].g[1,0] = Omega_b*(-R_s*i_sq + omega_s*psi_sd - v_h*cos(theta_h))
        struct[0].g[2,0] = L_m*i_rd + i_sd*(L_ls + L_m) - psi_sd
        struct[0].g[3,0] = L_m*i_rq + i_sq*(L_ls + L_m) - psi_sq
        struct[0].g[4,0] = L_m*i_sd + i_rd*(L_lr + L_m) - psi_rd
        struct[0].g[5,0] = L_m*i_sq + i_rq*(L_lr + L_m) - psi_rq
        struct[0].g[6,0] = P_h + v_0*v_h*sin(theta_0 - theta_h)/X_l
        struct[0].g[7,0] = Q_h + v_0*v_h*cos(theta_0 - theta_h)/X_l - v_h**2/X_l
        struct[0].g[8,0] = -P_h - i_sd*v_h*sin(theta_h) + i_sq*v_h*cos(theta_h)
        struct[0].g[9,0] = B_c*v_h**2 - Q_h - i_sd*v_h*cos(theta_h) - i_sq*v_h*sin(theta_h)
    
    # Outputs:
    if mode == 3:

        struct[0].h[0,0] = I_b*(i_sd**2 + i_sq**2)**0.5
        struct[0].h[1,0] = -v_h*sin(theta_h)
        struct[0].h[2,0] = v_h*cos(theta_h)
        struct[0].h[3,0] = B_c*v_h**2
    

    if mode == 10:

        struct[0].Fx[0,0] = -D/(2*H_m)
        struct[0].Fx[1,0] = Omega_b*psi_rq
        struct[0].Fx[1,2] = Omega_b*(omega_e - omega_s)
        struct[0].Fx[2,0] = -Omega_b*psi_rd
        struct[0].Fx[2,1] = Omega_b*(-omega_e + omega_s)

    if mode == 11:

        struct[0].Fy[0,0] = -i_sq/(2*H_m)
        struct[0].Fy[0,1] = i_sd/(2*H_m)
        struct[0].Fy[0,2] = psi_sq/(2*H_m)
        struct[0].Fy[0,3] = -psi_sd/(2*H_m)
        struct[0].Fy[1,4] = -Omega_b*R_r
        struct[0].Fy[2,5] = -Omega_b*R_r

        struct[0].Gx[4,1] = -1
        struct[0].Gx[5,2] = -1

        struct[0].Gy[0,1] = -Omega_b*omega_s
        struct[0].Gy[0,2] = -Omega_b*R_s
        struct[0].Gy[0,8] = Omega_b*sin(theta_h)
        struct[0].Gy[0,9] = Omega_b*v_h*cos(theta_h)
        struct[0].Gy[1,0] = Omega_b*omega_s
        struct[0].Gy[1,3] = -Omega_b*R_s
        struct[0].Gy[1,8] = -Omega_b*cos(theta_h)
        struct[0].Gy[1,9] = Omega_b*v_h*sin(theta_h)
        struct[0].Gy[2,0] = -1
        struct[0].Gy[2,2] = L_ls + L_m
        struct[0].Gy[2,4] = L_m
        struct[0].Gy[3,1] = -1
        struct[0].Gy[3,3] = L_ls + L_m
        struct[0].Gy[3,5] = L_m
        struct[0].Gy[4,2] = L_m
        struct[0].Gy[4,4] = L_lr + L_m
        struct[0].Gy[5,3] = L_m
        struct[0].Gy[5,5] = L_lr + L_m
        struct[0].Gy[6,6] = 1
        struct[0].Gy[6,8] = v_0*sin(theta_0 - theta_h)/X_l
        struct[0].Gy[6,9] = -v_0*v_h*cos(theta_0 - theta_h)/X_l
        struct[0].Gy[7,7] = 1
        struct[0].Gy[7,8] = v_0*cos(theta_0 - theta_h)/X_l - 2*v_h/X_l
        struct[0].Gy[7,9] = v_0*v_h*sin(theta_0 - theta_h)/X_l
        struct[0].Gy[8,2] = -v_h*sin(theta_h)
        struct[0].Gy[8,3] = v_h*cos(theta_h)
        struct[0].Gy[8,6] = -1
        struct[0].Gy[8,8] = -i_sd*sin(theta_h) + i_sq*cos(theta_h)
        struct[0].Gy[8,9] = -i_sd*v_h*cos(theta_h) - i_sq*v_h*sin(theta_h)
        struct[0].Gy[9,2] = -v_h*cos(theta_h)
        struct[0].Gy[9,3] = -v_h*sin(theta_h)
        struct[0].Gy[9,7] = -1
        struct[0].Gy[9,8] = 2*B_c*v_h - i_sd*cos(theta_h) - i_sq*sin(theta_h)
        struct[0].Gy[9,9] = i_sd*v_h*sin(theta_h) - i_sq*v_h*cos(theta_h)



@numba.njit(cache=True)
def ini(struct,mode):

    # Parameters:
    S_b = struct[0].S_b
    U_b = struct[0].U_b
    I_b = struct[0].I_b
    R_s = struct[0].R_s
    R_r = struct[0].R_r
    L_ls = struct[0].L_ls
    L_lr = struct[0].L_lr
    L_m = struct[0].L_m
    H_m = struct[0].H_m
    Omega_b = struct[0].Omega_b
    D = struct[0].D
    v_0 = struct[0].v_0
    theta_0 = struct[0].theta_0
    X_l = struct[0].X_l
    omega_s = struct[0].omega_s
    v_rd = struct[0].v_rd
    v_rq = struct[0].v_rq
    
    # Inputs:
    P_h = struct[0].P_h
    Q_h = struct[0].Q_h
    
    # Dynamical states:
    omega_e = struct[0].x_ini[0,0]
    psi_rd = struct[0].x_ini[1,0]
    psi_rq = struct[0].x_ini[2,0]
    
    # Algebraic states:
    psi_sd = struct[0].y_ini[0,0]
    psi_sq = struct[0].y_ini[1,0]
    i_sd = struct[0].y_ini[2,0]
    i_sq = struct[0].y_ini[3,0]
    i_rd = struct[0].y_ini[4,0]
    i_rq = struct[0].y_ini[5,0]
    tau_m = struct[0].y_ini[6,0]
    B_c = struct[0].y_ini[7,0]
    v_h = struct[0].y_ini[8,0]
    theta_h = struct[0].y_ini[9,0]
    
    # Differential equations:
    if mode == 2:


        struct[0].f_ini[0,0] = (-D*omega_e + i_sd*psi_sq - i_sq*psi_sd + tau_m)/(2*H_m)
        struct[0].f_ini[1,0] = Omega_b*(-R_r*i_rd - psi_rq*(-omega_e + omega_s) - v_rd)
        struct[0].f_ini[2,0] = Omega_b*(-R_r*i_rq + psi_rd*(-omega_e + omega_s) - v_rq)
    
    # Algebraic equations:
    if mode == 3:


        struct[0].g_ini[0,0] = Omega_b*(-R_s*i_sd - omega_s*psi_sq + v_h*sin(theta_h))
        struct[0].g_ini[1,0] = Omega_b*(-R_s*i_sq + omega_s*psi_sd - v_h*cos(theta_h))
        struct[0].g_ini[2,0] = L_m*i_rd + i_sd*(L_ls + L_m) - psi_sd
        struct[0].g_ini[3,0] = L_m*i_rq + i_sq*(L_ls + L_m) - psi_sq
        struct[0].g_ini[4,0] = L_m*i_sd + i_rd*(L_lr + L_m) - psi_rd
        struct[0].g_ini[5,0] = L_m*i_sq + i_rq*(L_lr + L_m) - psi_rq
        struct[0].g_ini[6,0] = P_h + v_0*v_h*sin(theta_0 - theta_h)/X_l
        struct[0].g_ini[7,0] = Q_h + v_0*v_h*cos(theta_0 - theta_h)/X_l - v_h**2/X_l
        struct[0].g_ini[8,0] = -P_h - i_sd*v_h*sin(theta_h) + i_sq*v_h*cos(theta_h)
        struct[0].g_ini[9,0] = B_c*v_h**2 - Q_h - i_sd*v_h*cos(theta_h) - i_sq*v_h*sin(theta_h)
    
    # Outputs:
    if mode == 3:

        struct[0].h[0,0] = I_b*(i_sd**2 + i_sq**2)**0.5
        struct[0].h[1,0] = -v_h*sin(theta_h)
        struct[0].h[2,0] = v_h*cos(theta_h)
        struct[0].h[3,0] = B_c*v_h**2
    

    if mode == 10:

        struct[0].Fx_ini[0,0] = -D/(2*H_m)
        struct[0].Fx_ini[1,0] = Omega_b*psi_rq
        struct[0].Fx_ini[1,2] = Omega_b*(omega_e - omega_s)
        struct[0].Fx_ini[2,0] = -Omega_b*psi_rd
        struct[0].Fx_ini[2,1] = Omega_b*(-omega_e + omega_s)

    if mode == 11:

        struct[0].Fy_ini[0,0] = -i_sq/(2*H_m)
        struct[0].Fy_ini[0,1] = i_sd/(2*H_m)
        struct[0].Fy_ini[0,2] = psi_sq/(2*H_m)
        struct[0].Fy_ini[0,3] = -psi_sd/(2*H_m)
        struct[0].Fy_ini[1,4] = -Omega_b*R_r
        struct[0].Fy_ini[2,5] = -Omega_b*R_r

        struct[0].Gx_ini[4,1] = -1
        struct[0].Gx_ini[5,2] = -1

        struct[0].Gy_ini[0,1] = -Omega_b*omega_s
        struct[0].Gy_ini[0,2] = -Omega_b*R_s
        struct[0].Gy_ini[0,8] = Omega_b*sin(theta_h)
        struct[0].Gy_ini[0,9] = Omega_b*v_h*cos(theta_h)
        struct[0].Gy_ini[1,0] = Omega_b*omega_s
        struct[0].Gy_ini[1,3] = -Omega_b*R_s
        struct[0].Gy_ini[1,8] = -Omega_b*cos(theta_h)
        struct[0].Gy_ini[1,9] = Omega_b*v_h*sin(theta_h)
        struct[0].Gy_ini[2,0] = -1
        struct[0].Gy_ini[2,2] = L_ls + L_m
        struct[0].Gy_ini[2,4] = L_m
        struct[0].Gy_ini[3,1] = -1
        struct[0].Gy_ini[3,3] = L_ls + L_m
        struct[0].Gy_ini[3,5] = L_m
        struct[0].Gy_ini[4,2] = L_m
        struct[0].Gy_ini[4,4] = L_lr + L_m
        struct[0].Gy_ini[5,3] = L_m
        struct[0].Gy_ini[5,5] = L_lr + L_m
        struct[0].Gy_ini[6,8] = v_0*sin(theta_0 - theta_h)/X_l
        struct[0].Gy_ini[6,9] = -v_0*v_h*cos(theta_0 - theta_h)/X_l
        struct[0].Gy_ini[7,8] = v_0*cos(theta_0 - theta_h)/X_l - 2*v_h/X_l
        struct[0].Gy_ini[7,9] = v_0*v_h*sin(theta_0 - theta_h)/X_l
        struct[0].Gy_ini[8,2] = -v_h*sin(theta_h)
        struct[0].Gy_ini[8,3] = v_h*cos(theta_h)
        struct[0].Gy_ini[8,8] = -i_sd*sin(theta_h) + i_sq*cos(theta_h)
        struct[0].Gy_ini[8,9] = -i_sd*v_h*cos(theta_h) - i_sq*v_h*sin(theta_h)
        struct[0].Gy_ini[9,2] = -v_h*cos(theta_h)
        struct[0].Gy_ini[9,3] = -v_h*sin(theta_h)
        struct[0].Gy_ini[9,7] = v_h**2
        struct[0].Gy_ini[9,8] = 2*B_c*v_h - i_sd*cos(theta_h) - i_sq*sin(theta_h)
        struct[0].Gy_ini[9,9] = i_sd*v_h*sin(theta_h) - i_sq*v_h*cos(theta_h)





@numba.njit(cache=True)
def Piecewise(arg):
    out = arg[0][1]
    N = len(arg)
    for it in range(N-1,-1,-1):
        if arg[it][1]: out = arg[it][0]
    return out

@numba.njit(cache=True)
def ITE(arg):
    out = arg[0][1]
    N = len(arg)
    for it in range(N-1,-1,-1):
        if arg[it][1]: out = arg[it][0]
    return out


@numba.njit(cache=True)
def Abs(x):
    return np.abs(x)



@numba.njit(cache=True) 
def daesolver(struct): 
    sin = np.sin
    cos = np.cos
    sqrt = np.sqrt
    i = 0 
    
    Dt = struct[i].Dt 

    N_x = struct[i].N_x
    N_y = struct[i].N_y
    N_z = struct[i].N_z

    decimation = struct[i].decimation 
    eye = np.eye(N_x)
    t = struct[i].t 
    t_end = struct[i].t_end 
    if struct[i].it == 0:
        run(t,struct, 1) 
        struct[i].it_store = 0  
        struct[i]['T'][0] = t 
        struct[i].X[0,:] = struct[i].x[:,0]  
        struct[i].Y[0,:] = struct[i].y[:,0]  
        struct[i].Z[0,:] = struct[i].h[:,0]  

    solver = struct[i].solvern 
    while t<t_end: 
        struct[i].it += 1
        struct[i].t += Dt
        
        it = struct[i].it
        t = struct[i].t
        it_store = struct[i].it_store
        
        #perturbations(t,struct) 
        
        if solver == 1: 
            # forward euler solver  
            run(t,struct, 2)  
            struct[i].x[:] += Dt*struct[i].f  
 
        if solver == 2: 
            
            # bacward euler solver
            x_0 = np.copy(struct[i].x[:]) 
            for j in range(struct[i].imax): 
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct, 10)  
                phi =  x_0 + Dt*struct[i].f - struct[i].x 
                Dx = np.linalg.solve(-(Dt*struct[i].Fx - np.eye(N_x)), phi) 
                struct[i].x[:] += Dx[:] 
                if np.max(np.abs(Dx)) < struct[i].itol: break 
            print(struct[i].f)
 
        if solver == 3: 
            # trapezoidal solver
            run(t,struct, 2) 
            f_0 = np.copy(struct[i].f[:]) 
            x_0 = np.copy(struct[i].x[:]) 
            for j in range(struct[i].imax): 
                run(t,struct, 10)  
                phi =  x_0 + 0.5*Dt*(f_0 + struct[i].f) - struct[i].x 
                Dx = np.linalg.solve(-(0.5*Dt*struct[i].Fx - np.eye(N_x)), phi) 
                struct[i].x[:] += Dx[:] 
                run(t,struct, 2) 
                if np.max(np.abs(Dx)) < struct[i].itol: break 

        if solver == 4: # Teapezoidal DAE as in Milano's book

            run(t,struct, 2) 
            run(t,struct, 3) 

            x = np.copy(struct[i].x[:]) 
            y = np.copy(struct[i].y[:]) 
            f = np.copy(struct[i].f[:]) 
            g = np.copy(struct[i].g[:]) 
            
            for iter in range(struct[i].imax):
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct,10) 
                run(t,struct,11) 
                
                x_i = struct[i].x[:] 
                y_i = struct[i].y[:]  
                f_i = struct[i].f[:] 
                g_i = struct[i].g[:]                 
                F_x_i = struct[i].Fx[:,:]
                F_y_i = struct[i].Fy[:,:] 
                G_x_i = struct[i].Gx[:,:] 
                G_y_i = struct[i].Gy[:,:]                

                A_c_i = np.vstack((np.hstack((eye-0.5*Dt*F_x_i, -0.5*Dt*F_y_i)),
                                   np.hstack((G_x_i,         G_y_i))))
                     
                f_n_i = x_i - x - 0.5*Dt*(f_i+f) 
                # print(t,iter,g_i)
                Dxy_i = np.linalg.solve(-A_c_i,np.vstack((f_n_i,g_i))) 
                
                x_i = x_i + Dxy_i[0:N_x]
                y_i = y_i + Dxy_i[N_x:(N_x+N_y)]

                struct[i].x[:] = x_i
                struct[i].y[:] = y_i

                if np.max(np.abs(Dxy_i[:,0]))<struct[i].itol:
                    
                    break
                
                # if iter>struct[i].imax-2:
                    
                #     print('Convergence problem')

            struct[i].x[:] = x_i
            struct[i].y[:] = y_i
            
        if solver == 5: # Teapezoidal DAE as in Milano's book

            run(t,struct, 2) 
            run(t,struct, 3) 

            x = np.copy(struct[i].x[:]) 
            y = np.copy(struct[i].y[:]) 
            f = np.copy(struct[i].f[:]) 
            g = np.copy(struct[i].g[:]) 
            
            for iter in range(struct[i].imax):
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct,10) 
                run(t,struct,11) 
                
                x_i = struct[i].x[:] 
                y_i = struct[i].y[:]  
                f_i = struct[i].f[:] 
                g_i = struct[i].g[:]                 
                F_x_i = struct[i].Fx[:,:]
                F_y_i = struct[i].Fy[:,:] 
                G_x_i = struct[i].Gx[:,:] 
                G_y_i = struct[i].Gy[:,:]                

                A_c_i = np.vstack((np.hstack((eye-0.5*Dt*F_x_i, -0.5*Dt*F_y_i)),
                                   np.hstack((G_x_i,         G_y_i))))
                     
                f_n_i = x_i - x - 0.5*Dt*(f_i+f) 
                # print(t,iter,g_i)
                Dxy_i = np.linalg.solve(-A_c_i,np.vstack((f_n_i,g_i))) 
                
                x_i = x_i + Dxy_i[0:N_x]
                y_i = y_i + Dxy_i[N_x:(N_x+N_y)]

                struct[i].x[:] = x_i
                struct[i].y[:] = y_i

        # [f_i,g_i,F_x_i,F_y_i,G_x_i,G_y_i] =  smib_transient(x_i,y_i,u);
        
        # A_c_i = [[eye(N_x)-0.5*Dt*F_x_i, -0.5*Dt*F_y_i],
        #          [                G_x_i,         G_y_i]];
             
        # f_n_i = x_i - x - 0.5*Dt*(f_i+f);
        
        # Dxy_i = -A_c_i\[f_n_i.',g_i.'].';
        
        # x_i = x_i + Dxy_i(1:N_x);
        # y_i = y_i + Dxy_i(N_x+1:N_x+N_y);
                
                xy = np.vstack((x_i,y_i))
                max_relative = 0.0
                for it_var in range(N_x+N_y):
                    abs_value = np.abs(xy[it_var,0])
                    if abs_value < 0.001:
                        abs_value = 0.001
                                             
                    relative_error = np.abs(Dxy_i[it_var,0])/abs_value
                    
                    if relative_error > max_relative: max_relative = relative_error
                    
                if max_relative<struct[i].itol:
                    
                    break
                
                # if iter>struct[i].imax-2:
                    
                #     print('Convergence problem')

            struct[i].x[:] = x_i
            struct[i].y[:] = y_i
                
        # channels 
        if it >= it_store*decimation: 
            struct[i]['T'][it_store+1] = t 
            struct[i].X[it_store+1,:] = struct[i].x[:,0] 
            struct[i].Y[it_store+1,:] = struct[i].y[:,0]
            struct[i].Z[it_store+1,:] = struct[i].h[:,0]
            struct[i].iters[it_store+1,0] = iter
            struct[i].it_store += 1 
            
    struct[i].t = t
    struct[i].it_store = it_store
    return t


import numpy as np
import sympy as sym
import pydae.build_cffi as db
from pydae.bmapu import bmapu_builder


data = {
"sys":{"name":"smib","S_base":100e6, "K_sec":0.01,"K_p_agc":1.0,"K_i_agc":0.1},       
"buses":[{"name":"1", "P_W":0.0,"Q_var":0.0,"U_kV":20.0},
         {"name":"2", "P_W":500.0e6,"Q_var":0.0,"U_kV":20.0}
        ],
"lines":[{"bus_j":"1", "bus_k":"2", "X_pu":0.05,"R_pu":0.01,"Bs_pu":1e-6,"S_mva":2e3}],
"syns":[
      {"bus":"1","S_n":900e6,
         "X_d":1.8,"X1d":0.3, "T1d0":8.0,    
         "X_q":1.7,"X1q":0.55,"T1q0":0.4,  
         "R_a":0.01,"X_l": 0.2, 
         "H":5.0,"D":1.0,
         "Omega_b":314.1592653589793,"omega_s":1.0,"K_sec":0.0,
         "gov":{"type":"agov1","Droop":0.05,"T_1":1.0,"T_2":2.0,"T_3":10.0, "p_c":0.01,"omega_ref":1.0, "K_imw":0.1},
         "K_delta":0.0},
      {"bus":"2","S_n":90000e6,
         "X_d":1.8,"X1d":0.3, "T1d0":8.0,    
         "X_q":1.7,"X1q":0.55,"T1q0":0.4,  
         "R_a":0.0025,"X_l": 0.2, 
         "H":5.0,"D":1.0,
         "Omega_b":314.1592653589793,"omega_s":1.0,"K_sec":0.01,
         "avr":{"type":"sexs","K_a":100, "T_r":0.02, "K_ai":1e-6, "v_pss":0.0, "v_ref":1.0,"V_min":-5,"V_max":5.0,"K_aw":2.0},
         "gov":{"type":"agov1","Droop":0.05,"T_1":1.0,"T_2":2.0,"T_3":10.0, "p_c":0.01,"omega_ref":1.0, "K_imw":0.0},
         "K_delta":0.01}
         ]
}

grid = bmapu_builder.bmapu(data)
grid.checker()
grid.build()

g_list = grid.dae['g'] 
h_dict = grid.dae['h_dict']
f_list = grid.dae['f']
x_list = grid.dae['x']
params_dict = grid.dae['params_dict']


sys = {'name':'smib',
       'params_dict':params_dict,
       'f_list':f_list,
       'g_list':g_list,
       'x_list':x_list,
       'y_ini_list':grid.dae['y_ini'],
       'y_run_list':grid.dae['y_run'],
       'u_run_dict':grid.dae['u_run_dict'],
       'u_ini_dict':grid.dae['u_ini_dict'],
       'h_dict':h_dict}

dblr = db.builder(sys)
dblr.build()

import numpy as np
import matplotlib.pyplot as plt
import ipywidgets
plt.style.use(r'https://raw.githubusercontent.com/jmmauricio/edashboards/main/presentation.mplstyle')
colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

import smib
grid = smib.model()


class dashboard(smib.model):
    
    def __init__(self):
        
        super().__init__()
        
        grid = smib.model()
        grid.Dt = 0.01
        grid.decimation = 1
        grid.ini({'p_c_1':0.0,'v_f_1':1.0,
                  'P_2':-50000e6,'D_1':0,
                 "T_gov_1_1":1.0,"T_gov_2_1":1.0,"T_gov_3_1":1.0, "K_imw_1":1,"Droop_1":1e3},'xy_0.json')
        grid.run(30.0,{})
        grid.post();
        
        self.grid = grid
        
        plt.style.use(r'https://raw.githubusercontent.com/jmmauricio/edashboards/main/presentation.mplstyle')
        self.colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        self.widgets()
        
    def widgets(self):
        
        grid = self.grid
        colors = self.colors

        plt.ioff()
        plt.clf()


        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(9, 4), frameon=False)
        fig.canvas.toolbar_visible = False

        self.line_delta = axes[0,0].plot(grid.Time, grid.get_values('delta_1'), label='$\sf \delta$', color=colors[4])
        self.line_omega = axes[1,0].plot(grid.Time, grid.get_values('omega_1'), label='$\sf \omega$', color=colors[1])
        self.line_v_1 = axes[0,1].plot(grid.Time, grid.get_values('V_1'), label='$\sf V_1$', color=colors[5])
        #line_theta_1 = axes[0,1].plot(T, Y[:,syst.y_list.index('theta_1')], label='$\sf \\theta_1$')
        self.line_p_t = axes[1,1].plot(grid.Time, grid.get_values('p_g_1'), label='$\sf P_t$', color=colors[2])
        self.line_q_t = axes[1,1].plot(grid.Time, grid.get_values('q_g_1'), label='$\sf Q_t$', color=colors[0])

        y_labels = ['$\delta$','$\omega$','$P_t$']

        axes[0,0].set_ylim((-1,2))
        axes[1,0].set_ylim((0.95,1.05))
        axes[0,1].set_ylim((0.8,1.2))
        axes[1,1].set_ylim((-0.5,1.5))

        axes[0,0].grid(True)
        axes[1,0].grid(True)
        axes[0,1].grid(True)
        axes[1,1].grid(True)
        axes[0,0].legend(loc='best')
        axes[1,0].legend(loc='best')
        axes[0,1].legend(loc='best')
        axes[1,1].legend(loc='best')

        axes[1,0].set_xlabel('Time (s)')  
        axes[1,1].set_xlabel('Time (s)') 

        fig.tight_layout()
        
        self.fig = fig
        #axes[0].set_title('Par en función de la velocidad')
        #axes[1].set_title('Corriente en función de la velocidad')


        self.sld_p_m = ipywidgets.FloatSlider(orientation='horizontal',
                                        description_allow_html=True,description="p<sub>m</sub>", 
                                        value=grid.get_value('p_m_1'), min=0.0,max= 1.2, 
                                        step=.1)


        self.sld_v_f = ipywidgets.FloatSlider(orientation='horizontal',
                                        description_allow_html=True,description="v<sub>f</sub>", 
                                        value=grid.get_value('v_f_1'), min=0.5,max= 4, 
                                        step=.1)

        self.prog_c = ipywidgets.IntProgress(
            value=100,
            min=0,
            max=120,
            step=1,
            description='SM Load:',
            bar_style='', # 'success', 'info', 'warning', 'danger' or ''
            orientation='horizontal' 
        )

        self.prog_damp = ipywidgets.IntProgress(
            value=10,
            min=0,
            max=20,
            step=1,
            description='ζ = 1.0',
            bar_style='', # 'success', 'info', 'warning', 'danger' or ''
            orientation='horizontal' 
        )
        
    def update(self,change):
        
        grid = self.grid

        p_m = self.sld_p_m.value
        v_f = self.sld_v_f.value


        #grid = smib.smib_class()
        #grid.Dt = 0.01
        grid.decimation = 10
        grid.Dt = 0.01
        grid.ini({'p_c_1':0.0,'v_f_1':1.0,
                  'P_2':-50000e6,'D_1':0,
                 "T_gov_1_1":0.1,"T_gov_2_1":1.0,"T_gov_3_1":1.0, "K_imw_1":1,"Droop_1":1e3},'xy_0.json')
        grid.run( 1.0,{})
        grid.run(10,{'p_c_1':p_m,'v_f_1':v_f})
        grid.Dt = 0.1
        grid.run(30,{'p_c_1':p_m,'v_f_1':v_f})

        grid.post();


        self.line_delta[0].set_data(grid.Time, grid.get_values('delta_1'))
        self.line_omega[0].set_data(grid.Time, grid.get_values('omega_1'))
        self.line_v_1[0].set_data(grid.Time, grid.get_values('V_1'))
        #line_theta_1 = axes[0,1].plot(T, Y[:,syst.y_list.index('theta_1')], label='$\sf \\theta_1$')
        self.line_p_t[0].set_data(grid.Time, grid.get_values('p_g_1'))
        self.line_q_t[0].set_data(grid.Time, grid.get_values('q_g_1'))

        i_d, i_q = grid.get_mvalue(['i_d_1','i_q_1'])
        c = (i_d**2+i_q**2)*0.5

        self.prog_c.bar_style = 'success'
        if c>0.9:
            self.prog_c.bar_style = 'warning'
        if c>1.0:
            self.prog_c.bar_style = 'danger'
        self.prog_c.value = 100*c

        self.fig.canvas.draw_idle()


    def show(self):

        self.sld_p_m.observe(self.update, names='value')
        self.sld_v_f.observe(self.update, names='value')

        layout_row1 = ipywidgets.HBox([self.fig.canvas])
        layout_row2 = ipywidgets.HBox([self.sld_p_m,self.sld_v_f,self.prog_c])

        layout = ipywidgets.VBox([layout_row1,layout_row2])
        self.layout = layout
        display(self.layout)
     


import numpy as np
import numba
import scipy.optimize as sopt


sin = np.sin
cos = np.cos

class smib_milano_ex8p1_4ord_avr_class: 

    def __init__(self): 

        self.t_end = 10.000000 
        self.Dt = 0.0010000 
        self.decimation = 10.000000 
        self.itol = 1e-6 
        self.Dt_max = 0.001000 
        self.Dt_min = 0.001000 
        self.solvern = 5 
        self.imax = 100 
        self.N_x = 5
        self.N_y = 7 
        self.N_z = 1 
        self.N_store = 10000 
        self.params_list = ['X_d', 'X1d', 'T1d0', 'X_q', 'X1q', 'T1q0', 'R_a', 'X_l', 'H', 'D', 'Omega_b', 'omega_s', 'v_0', 'theta_0', 'K_a', 'T_r', 'v_pss'] 
        self.params_values_list  = [1.81, 0.3, 8.0, 1.76, 0.65, 1.0, 0.003, 0.05, 3.5, 1.0, 314.1592653589793, 1.0, 1.0, 0.0, 100, 0.1, 0.0] 
        self.inputs_ini_list = ['P_t', 'Q_t'] 
        self.inputs_ini_values_list  = [0.8, 0.2] 
        self.inputs_run_list = ['p_m', 'v_ref'] 
        self.inputs_run_values_list = [0.8, 1.0] 
        self.x_list = ['delta', 'omega', 'e1q', 'e1d', 'v_c'] 
        self.y_list = ['i_d', 'i_q', 'v_1', 'theta_1', 'P_t', 'Q_t', 'v_f'] 
        self.xy_list = self.x_list + self.y_list 
        self.y_ini_list = ['i_d', 'i_q', 'v_1', 'theta_1', 'p_m', 'v_ref', 'v_f'] 
        self.xy_ini_list = self.x_list + self.y_ini_list 
        self.t = 0.0
        self.it = 0
        self.it_store = 0
        self.update() 

    def update(self): 

        self.N_steps = int(np.ceil(self.t_end/self.Dt)) 
        dt = [  
              ('t_end', np.float64),
              ('Dt', np.float64),
              ('decimation', np.float64),
              ('itol', np.float64),
              ('Dt_max', np.float64),
              ('Dt_min', np.float64),
              ('solvern', np.int64),
              ('imax', np.int64),
              ('N_steps', np.int64),
              ('N_store', np.int64),
              ('N_x', np.int64),
              ('N_y', np.int64),
              ('t', np.float64),
              ('it', np.int64),
              ('it_store', np.int64),
              ('idx', np.int64),
              ('idy', np.int64),
              ('f', np.float64, (self.N_x,1)),
              ('f_ini', np.float64, (self.N_x,1)),
              ('x', np.float64, (self.N_x,1)),
              ('x_ini', np.float64, (self.N_x,1)),
              ('x_0', np.float64, (self.N_x,1)),
              ('g', np.float64, (self.N_y,1)),
              ('g_ini', np.float64, (self.N_y,1)),
              ('y', np.float64, (self.N_y,1)),
              ('y_ini', np.float64, (self.N_y,1)),
              ('y_0', np.float64, (self.N_y,1)),
              ('h', np.float64, (self.N_z,1)),
              ('Fx', np.float64, (self.N_x,self.N_x)),
              ('Fy', np.float64, (self.N_x,self.N_y)),
              ('Gx', np.float64, (self.N_y,self.N_x)),
              ('Gy', np.float64, (self.N_y,self.N_y)),
              ('Fx_ini', np.float64, (self.N_x,self.N_x)),
              ('Fy_ini', np.float64, (self.N_x,self.N_y)),
              ('Gx_ini', np.float64, (self.N_y,self.N_x)),
              ('Gy_ini', np.float64, (self.N_y,self.N_y)),
              ('T', np.float64, (self.N_store+1,1)),
              ('X', np.float64, (self.N_store+1,5)),
              ('Y', np.float64, (self.N_store+1,7)),
              ('iters', np.float64, (self.N_store+1,1)),
             ]

        values = [
                self.t_end,                          
                self.Dt,
                self.decimation,
                self.itol,
                self.Dt_max,
                self.Dt_min,
                self.solvern,
                self.imax,
                self.N_steps,
                self.N_store,
                self.N_x,
                self.N_y,
                self.t,
                self.it,
                self.it_store,
                0,                                     # idx
                0,                                     # idy
                np.zeros((self.N_x,1)),                # f
                np.zeros((self.N_x,1)),                # f_ini
                np.zeros((self.N_x,1)),                # x
                np.zeros((self.N_x,1)),                # x_ini
                np.zeros((self.N_x,1)),                # x_0
                np.zeros((self.N_y,1)),                # g
                np.zeros((self.N_y,1)),                # g_ini
                np.zeros((self.N_y,1)),                # y
                np.zeros((self.N_y,1)),                # y_ini
                np.zeros((self.N_y,1)),                # y_0
                np.zeros((self.N_z,1)),                # h
                np.zeros((self.N_x,self.N_x)),         # Fx   
                np.zeros((self.N_x,self.N_y)),         # Fy 
                np.zeros((self.N_y,self.N_x)),         # Gx 
                np.zeros((self.N_y,self.N_y)),         # Fy 
                np.zeros((self.N_x,self.N_x)),         # Fx_ini  
                np.zeros((self.N_x,self.N_y)),         # Fy_ini 
                np.zeros((self.N_y,self.N_x)),         # Gx_ini 
                np.zeros((self.N_y,self.N_y)),         # Fy_ini 
                np.zeros((self.N_store+1,1)),          # T
                np.zeros((self.N_store+1,self.N_x)),   # X
                np.zeros((self.N_store+1,self.N_y)),   # Y
                np.zeros((self.N_store+1,1)),          # iters
                ]  

        dt += [(item,np.float64) for item in self.params_list]
        values += [item for item in self.params_values_list]

        for item_id,item_val in zip(self.inputs_ini_list,self.inputs_ini_values_list):
            if item_id in self.inputs_run_list: continue
            dt += [(item_id,np.float64)]
            values += [item_val]

        dt += [(item,np.float64) for item in self.inputs_run_list]
        values += [item for item in self.inputs_run_values_list]

        self.struct = np.rec.array([tuple(values)], dtype=np.dtype(dt))


    def ini_problem(self,x):
        self.struct[0].x_ini[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        ini(self.struct,2)
        ini(self.struct,3)       
        fg = np.vstack((self.struct[0].f_ini,self.struct[0].g_ini))[:,0]
        return fg

    def dae_jacobian(self,x):
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        run(0.0,self.struct,10)
        run(0.0,self.struct,11)       
        A_c = np.block([[self.struct[0].Fx,self.struct[0].Fy],
                        [self.struct[0].Gx,self.struct[0].Gy]])
        return A_c

    def ini_dae_jacobian(self,x):
        self.struct[0].x_ini[:,0] = x[0:self.N_x]
        self.struct[0].y_ini[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        ini(self.struct,10)
        ini(self.struct,11)       
        A_c = np.block([[self.struct[0].Fx_ini,self.struct[0].Fy_ini],
                        [self.struct[0].Gx_ini,self.struct[0].Gy_ini]])
        return A_c

    def run_problem(self,x):
        t = self.struct[0].t
        self.struct[0].x[:,0] = x[0:self.N_x]
        self.struct[0].y[:,0] = x[self.N_x:(self.N_x+self.N_y)]
        run(t,self.struct,2)
        run(t,self.struct,3)
        run(t,self.struct,10)
        run(t,self.struct,11)
        fg = np.vstack((self.struct[0].f,self.struct[0].g))[:,0]
        return fg

    def f_ode(self,x):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def f_odeint(self,x,t):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def f_ivp(self,t,x):
        self.struct[0].x[:,0] = x
        run(self.struct,1)
        return self.struct[0].f[:,0]

    def Fx_ode(self,x):
        self.struct[0].x[:,0] = x
        run(self.struct,10)
        return self.struct[0].Fx

    def eval_A(self):
        
        Fx = self.struct[0].Fx
        Fy = self.struct[0].Fy
        Gx = self.struct[0].Gx
        Gy = self.struct[0].Gy
        
        A = Fx - Fy @ np.linalg.solve(Gy,Gx)
        
        self.A = A
        
        return A

    def eval_A_ini(self):
        
        Fx = self.struct[0].Fx_ini
        Fy = self.struct[0].Fy_ini
        Gx = self.struct[0].Gx_ini
        Gy = self.struct[0].Gy_ini
        
        A = Fx - Fy @ np.linalg.solve(Gy,Gx)
        
        
        return A

    def simulate(self,events):
        # simulation parameters
        self.struct[0].it = 0       # set time step to zero
        self.struct[0].it_store = 0 # set storage to zero
        self.struct[0].t = 0.0      # set time to zero
                    
        # initialization
        it_event = 0
        event = events[it_event]
        for item in event:
            self.struct[0][item] = event[item]
            
        
        ## compute initial conditions using x and y_ini 
        xy0 = np.ones(self.N_x+self.N_y)
        xy = sopt.fsolve(self.ini_problem,xy0 )
        self.struct[0].x[:,0] = xy[0:self.N_x]
        self.struct[0].y[:,0] = xy[self.N_x:]

        ## y_ini to u_run
        for item in self.inputs_run_list:
            if item in self.y_ini_list:
                self.struct[0][item] = self.struct[0].y_ini[self.y_ini_list.index(item)]

        ## u_ini to y_run
        for item in self.inputs_ini_list:
            if item in self.y_list:
                self.struct[0].y[self.y_list.index(item)] = self.struct[0][item]
    
        ## solve selfem
        daesolver(self.struct)    # run until first event

        # simulation run
        for event in events[1:]:  
            for item in event:
                self.struct[0][item] = event[item]
            daesolver(self.struct)    # run until next event
            
        
        # post process result    
        T = self.struct[0]['T'][:self.struct[0].it_store]
        X = self.struct[0]['X'][:self.struct[0].it_store,:]
        Y = self.struct[0]['Y'][:self.struct[0].it_store,:]

        self.T = T
        self.X = X
        self.Y = Y

        return T,X,Y


@numba.njit(cache=True)
def run(t,struct,mode):

    # Parameters:
    X_d = struct[0].X_d
    X1d = struct[0].X1d
    T1d0 = struct[0].T1d0
    X_q = struct[0].X_q
    X1q = struct[0].X1q
    T1q0 = struct[0].T1q0
    R_a = struct[0].R_a
    X_l = struct[0].X_l
    H = struct[0].H
    D = struct[0].D
    Omega_b = struct[0].Omega_b
    omega_s = struct[0].omega_s
    v_0 = struct[0].v_0
    theta_0 = struct[0].theta_0
    K_a = struct[0].K_a
    T_r = struct[0].T_r
    v_pss = struct[0].v_pss
    
    # Inputs:
    p_m = struct[0].p_m
    v_ref = struct[0].v_ref
    
    # Dynamical states:
    delta = struct[0].x[0,0]
    omega = struct[0].x[1,0]
    e1q = struct[0].x[2,0]
    e1d = struct[0].x[3,0]
    v_c = struct[0].x[4,0]
    
    # Algebraic states:
    i_d = struct[0].y[0,0]
    i_q = struct[0].y[1,0]
    v_1 = struct[0].y[2,0]
    theta_1 = struct[0].y[3,0]
    P_t = struct[0].y[4,0]
    Q_t = struct[0].y[5,0]
    v_f = struct[0].y[6,0]
    
    # Differential equations:
    if mode == 2:


        struct[0].f[0,0] = Omega_b*(omega - omega_s)
        struct[0].f[1,0] = (-D*(omega - omega_s) - i_d*(R_a*i_d + v_1*sin(delta - theta_1)) - i_q*(R_a*i_q + v_1*cos(delta - theta_1)) + p_m)/(2*H)
        struct[0].f[2,0] = (-e1q - i_d*(-X1d + X_d) + v_f)/T1d0
        struct[0].f[3,0] = (-e1d + i_q*(-X1q + X_q))/T1q0
        struct[0].f[4,0] = (v_1 - v_c)/T_r
    
    # Algebraic equations:
    if mode == 3:


        struct[0].g[0,0] = R_a*i_q + X1d*i_d - e1q + v_1*cos(delta - theta_1)
        struct[0].g[1,0] = R_a*i_d - X1q*i_q - e1d + v_1*sin(delta - theta_1)
        struct[0].g[2,0] = P_t + v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g[3,0] = Q_t + v_0*v_1*cos(theta_0 - theta_1)/X_l - v_1**2/X_l
        struct[0].g[4,0] = -P_t + i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].g[5,0] = -Q_t + i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].g[6,0] = K_a*(-v_c + v_pss + v_ref) - v_f
    

    if mode == 10:

        struct[0].Fx[0,1] = Omega_b
        struct[0].Fx[1,0] = (-i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fx[1,1] = -D/(2*H)
        struct[0].Fx[2,2] = -1/T1d0
        struct[0].Fx[3,3] = -1/T1q0
        struct[0].Fx[4,4] = -1/T_r

    if mode == 11:

        struct[0].Fy[1,0] = (-2*R_a*i_d - v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fy[1,1] = (-2*R_a*i_q - v_1*cos(delta - theta_1))/(2*H)
        struct[0].Fy[1,2] = (-i_d*sin(delta - theta_1) - i_q*cos(delta - theta_1))/(2*H)
        struct[0].Fy[1,3] = (i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fy[2,0] = (X1d - X_d)/T1d0
        struct[0].Fy[2,6] = 1/T1d0
        struct[0].Fy[3,1] = (-X1q + X_q)/T1q0
        struct[0].Fy[4,2] = 1/T_r

        struct[0].Gx[0,0] = -v_1*sin(delta - theta_1)
        struct[0].Gx[0,2] = -1
        struct[0].Gx[1,0] = v_1*cos(delta - theta_1)
        struct[0].Gx[1,3] = -1
        struct[0].Gx[4,0] = i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].Gx[5,0] = -i_d*v_1*sin(delta - theta_1) - i_q*v_1*cos(delta - theta_1)
        struct[0].Gx[6,4] = -K_a

        struct[0].Gy[0,0] = X1d
        struct[0].Gy[0,1] = R_a
        struct[0].Gy[0,2] = cos(delta - theta_1)
        struct[0].Gy[0,3] = v_1*sin(delta - theta_1)
        struct[0].Gy[1,0] = R_a
        struct[0].Gy[1,1] = -X1q
        struct[0].Gy[1,2] = sin(delta - theta_1)
        struct[0].Gy[1,3] = -v_1*cos(delta - theta_1)
        struct[0].Gy[2,2] = v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[2,3] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy[2,4] = 1
        struct[0].Gy[3,2] = v_0*cos(theta_0 - theta_1)/X_l - 2*v_1/X_l
        struct[0].Gy[3,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy[3,5] = 1
        struct[0].Gy[4,0] = v_1*sin(delta - theta_1)
        struct[0].Gy[4,1] = v_1*cos(delta - theta_1)
        struct[0].Gy[4,2] = i_d*sin(delta - theta_1) + i_q*cos(delta - theta_1)
        struct[0].Gy[4,3] = -i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1)
        struct[0].Gy[4,4] = -1
        struct[0].Gy[5,0] = v_1*cos(delta - theta_1)
        struct[0].Gy[5,1] = -v_1*sin(delta - theta_1)
        struct[0].Gy[5,2] = i_d*cos(delta - theta_1) - i_q*sin(delta - theta_1)
        struct[0].Gy[5,3] = i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].Gy[5,5] = -1
        struct[0].Gy[6,6] = -1



@numba.njit(cache=True)
def ini(struct,mode):

    # Parameters:
    X_d = struct[0].X_d
    X1d = struct[0].X1d
    T1d0 = struct[0].T1d0
    X_q = struct[0].X_q
    X1q = struct[0].X1q
    T1q0 = struct[0].T1q0
    R_a = struct[0].R_a
    X_l = struct[0].X_l
    H = struct[0].H
    D = struct[0].D
    Omega_b = struct[0].Omega_b
    omega_s = struct[0].omega_s
    v_0 = struct[0].v_0
    theta_0 = struct[0].theta_0
    K_a = struct[0].K_a
    T_r = struct[0].T_r
    v_pss = struct[0].v_pss
    
    # Inputs:
    P_t = struct[0].P_t
    Q_t = struct[0].Q_t
    
    # Dynamical states:
    delta = struct[0].x_ini[0,0]
    omega = struct[0].x_ini[1,0]
    e1q = struct[0].x_ini[2,0]
    e1d = struct[0].x_ini[3,0]
    v_c = struct[0].x_ini[4,0]
    
    # Algebraic states:
    i_d = struct[0].y_ini[0,0]
    i_q = struct[0].y_ini[1,0]
    v_1 = struct[0].y_ini[2,0]
    theta_1 = struct[0].y_ini[3,0]
    p_m = struct[0].y_ini[4,0]
    v_ref = struct[0].y_ini[5,0]
    v_f = struct[0].y_ini[6,0]
    
    # Differential equations:
    if mode == 2:


        struct[0].f_ini[0,0] = Omega_b*(omega - omega_s)
        struct[0].f_ini[1,0] = (-D*(omega - omega_s) - i_d*(R_a*i_d + v_1*sin(delta - theta_1)) - i_q*(R_a*i_q + v_1*cos(delta - theta_1)) + p_m)/(2*H)
        struct[0].f_ini[2,0] = (-e1q - i_d*(-X1d + X_d) + v_f)/T1d0
        struct[0].f_ini[3,0] = (-e1d + i_q*(-X1q + X_q))/T1q0
        struct[0].f_ini[4,0] = (v_1 - v_c)/T_r
    
    # Algebraic equations:
    if mode == 3:


        struct[0].g_ini[0,0] = R_a*i_q + X1d*i_d - e1q + v_1*cos(delta - theta_1)
        struct[0].g_ini[1,0] = R_a*i_d - X1q*i_q - e1d + v_1*sin(delta - theta_1)
        struct[0].g_ini[2,0] = P_t + v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].g_ini[3,0] = Q_t + v_0*v_1*cos(theta_0 - theta_1)/X_l - v_1**2/X_l
        struct[0].g_ini[4,0] = -P_t + i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].g_ini[5,0] = -Q_t + i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].g_ini[6,0] = K_a*(-v_c + v_pss + v_ref) - v_f
    

    if mode == 10:

        struct[0].Fx_ini[0,1] = Omega_b
        struct[0].Fx_ini[1,0] = (-i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fx_ini[1,1] = -D/(2*H)
        struct[0].Fx_ini[2,2] = -1/T1d0
        struct[0].Fx_ini[3,3] = -1/T1q0
        struct[0].Fx_ini[4,4] = -1/T_r

    if mode == 11:

        struct[0].Fy_ini[1,0] = (-2*R_a*i_d - v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fy_ini[1,1] = (-2*R_a*i_q - v_1*cos(delta - theta_1))/(2*H)
        struct[0].Fy_ini[1,2] = (-i_d*sin(delta - theta_1) - i_q*cos(delta - theta_1))/(2*H)
        struct[0].Fy_ini[1,3] = (i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1))/(2*H)
        struct[0].Fy_ini[2,0] = (X1d - X_d)/T1d0
        struct[0].Fy_ini[2,6] = 1/T1d0
        struct[0].Fy_ini[3,1] = (-X1q + X_q)/T1q0
        struct[0].Fy_ini[4,2] = 1/T_r

        struct[0].Gx_ini[0,0] = -v_1*sin(delta - theta_1)
        struct[0].Gx_ini[0,2] = -1
        struct[0].Gx_ini[1,0] = v_1*cos(delta - theta_1)
        struct[0].Gx_ini[1,3] = -1
        struct[0].Gx_ini[4,0] = i_d*v_1*cos(delta - theta_1) - i_q*v_1*sin(delta - theta_1)
        struct[0].Gx_ini[5,0] = -i_d*v_1*sin(delta - theta_1) - i_q*v_1*cos(delta - theta_1)
        struct[0].Gx_ini[6,4] = -K_a

        struct[0].Gy_ini[0,0] = X1d
        struct[0].Gy_ini[0,1] = R_a
        struct[0].Gy_ini[0,2] = cos(delta - theta_1)
        struct[0].Gy_ini[0,3] = v_1*sin(delta - theta_1)
        struct[0].Gy_ini[1,0] = R_a
        struct[0].Gy_ini[1,1] = -X1q
        struct[0].Gy_ini[1,2] = sin(delta - theta_1)
        struct[0].Gy_ini[1,3] = -v_1*cos(delta - theta_1)
        struct[0].Gy_ini[2,2] = v_0*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[2,3] = -v_0*v_1*cos(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[3,2] = v_0*cos(theta_0 - theta_1)/X_l - 2*v_1/X_l
        struct[0].Gy_ini[3,3] = v_0*v_1*sin(theta_0 - theta_1)/X_l
        struct[0].Gy_ini[4,0] = v_1*sin(delta - theta_1)
        struct[0].Gy_ini[4,1] = v_1*cos(delta - theta_1)
        struct[0].Gy_ini[4,2] = i_d*sin(delta - theta_1) + i_q*cos(delta - theta_1)
        struct[0].Gy_ini[4,3] = -i_d*v_1*cos(delta - theta_1) + i_q*v_1*sin(delta - theta_1)
        struct[0].Gy_ini[5,0] = v_1*cos(delta - theta_1)
        struct[0].Gy_ini[5,1] = -v_1*sin(delta - theta_1)
        struct[0].Gy_ini[5,2] = i_d*cos(delta - theta_1) - i_q*sin(delta - theta_1)
        struct[0].Gy_ini[5,3] = i_d*v_1*sin(delta - theta_1) + i_q*v_1*cos(delta - theta_1)
        struct[0].Gy_ini[6,5] = K_a
        struct[0].Gy_ini[6,6] = -1





@numba.njit(cache=True)
def Piecewise(arg):
    out = arg[0][1]
    N = len(arg)
    for it in range(N-1,-1,-1):
        if arg[it][1]: out = arg[it][0]
    return out


@numba.njit(cache=True) 
def daesolver(struct): 
    sin = np.sin
    cos = np.cos
    sqrt = np.sqrt
    i = 0 
    
    Dt = struct[i].Dt 

    N_x = struct[i].N_x
    N_y = struct[i].N_y

    decimation = struct[i].decimation 
    eye = np.eye(N_x)
    t = struct[i].t 
    t_end = struct[i].t_end 
    if struct[i].it == 0:
        run(t,struct, 1) 
        struct[i].it_store = 0  
        struct[i]['T'][0] = t 
        struct[i].X[0,:] = struct[i].x[:,0]  
        struct[i].Y[0,:] = struct[i].y[:,0]  


    solver = struct[i].solvern 
    while t<t_end: 
        struct[i].it += 1
        struct[i].t += Dt
        
        it = struct[i].it
        t = struct[i].t
        it_store = struct[i].it_store
        
        #perturbations(t,struct) 
        
        if solver == 1: 
            # forward euler solver  
            run(t,struct, 2)  
            struct[i].x[:] += Dt*struct[i].f  
 
        if solver == 2: 
            
            # bacward euler solver
            x_0 = np.copy(struct[i].x[:]) 
            for j in range(struct[i].imax): 
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct, 10)  
                phi =  x_0 + Dt*struct[i].f - struct[i].x 
                Dx = np.linalg.solve(-(Dt*struct[i].Fx - np.eye(N_x)), phi) 
                struct[i].x[:] += Dx[:] 
                if np.max(np.abs(Dx)) < struct[i].itol: break 
            print(struct[i].f)
 
        if solver == 3: 
            # trapezoidal solver
            run(t,struct, 2) 
            f_0 = np.copy(struct[i].f[:]) 
            x_0 = np.copy(struct[i].x[:]) 
            for j in range(struct[i].imax): 
                run(t,struct, 10)  
                phi =  x_0 + 0.5*Dt*(f_0 + struct[i].f) - struct[i].x 
                Dx = np.linalg.solve(-(0.5*Dt*struct[i].Fx - np.eye(N_x)), phi) 
                struct[i].x[:] += Dx[:] 
                run(t,struct, 2) 
                if np.max(np.abs(Dx)) < struct[i].itol: break 

        if solver == 4: # Teapezoidal DAE as in Milano's book

            run(t,struct, 2) 
            run(t,struct, 3) 

            x = np.copy(struct[i].x[:]) 
            y = np.copy(struct[i].y[:]) 
            f = np.copy(struct[i].f[:]) 
            g = np.copy(struct[i].g[:]) 
            
            for iter in range(struct[i].imax):
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct,10) 
                run(t,struct,11) 
                
                x_i = struct[i].x[:] 
                y_i = struct[i].y[:]  
                f_i = struct[i].f[:] 
                g_i = struct[i].g[:]                 
                F_x_i = struct[i].Fx[:,:]
                F_y_i = struct[i].Fy[:,:] 
                G_x_i = struct[i].Gx[:,:] 
                G_y_i = struct[i].Gy[:,:]                

                A_c_i = np.vstack((np.hstack((eye-0.5*Dt*F_x_i, -0.5*Dt*F_y_i)),
                                   np.hstack((G_x_i,         G_y_i))))
                     
                f_n_i = x_i - x - 0.5*Dt*(f_i+f) 
                # print(t,iter,g_i)
                Dxy_i = np.linalg.solve(-A_c_i,np.vstack((f_n_i,g_i))) 
                
                x_i = x_i + Dxy_i[0:N_x]
                y_i = y_i + Dxy_i[N_x:(N_x+N_y)]

                struct[i].x[:] = x_i
                struct[i].y[:] = y_i

                if np.max(np.abs(Dxy_i[:,0]))<struct[i].itol:
                    
                    break
                
                # if iter>struct[i].imax-2:
                    
                #     print('Convergence problem')

            struct[i].x[:] = x_i
            struct[i].y[:] = y_i
            
        if solver == 5: # Teapezoidal DAE as in Milano's book

            run(t,struct, 2) 
            run(t,struct, 3) 

            x = np.copy(struct[i].x[:]) 
            y = np.copy(struct[i].y[:]) 
            f = np.copy(struct[i].f[:]) 
            g = np.copy(struct[i].g[:]) 
            
            for iter in range(struct[i].imax):
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct,10) 
                run(t,struct,11) 
                
                x_i = struct[i].x[:] 
                y_i = struct[i].y[:]  
                f_i = struct[i].f[:] 
                g_i = struct[i].g[:]                 
                F_x_i = struct[i].Fx[:,:]
                F_y_i = struct[i].Fy[:,:] 
                G_x_i = struct[i].Gx[:,:] 
                G_y_i = struct[i].Gy[:,:]                

                A_c_i = np.vstack((np.hstack((eye-0.5*Dt*F_x_i, -0.5*Dt*F_y_i)),
                                   np.hstack((G_x_i,         G_y_i))))
                     
                f_n_i = x_i - x - 0.5*Dt*(f_i+f) 
                # print(t,iter,g_i)
                Dxy_i = np.linalg.solve(-A_c_i,np.vstack((f_n_i,g_i))) 
                
                x_i = x_i + Dxy_i[0:N_x]
                y_i = y_i + Dxy_i[N_x:(N_x+N_y)]

                struct[i].x[:] = x_i
                struct[i].y[:] = y_i

        # [f_i,g_i,F_x_i,F_y_i,G_x_i,G_y_i] =  smib_transient(x_i,y_i,u);
        
        # A_c_i = [[eye(N_x)-0.5*Dt*F_x_i, -0.5*Dt*F_y_i],
        #          [                G_x_i,         G_y_i]];
             
        # f_n_i = x_i - x - 0.5*Dt*(f_i+f);
        
        # Dxy_i = -A_c_i\[f_n_i.',g_i.'].';
        
        # x_i = x_i + Dxy_i(1:N_x);
        # y_i = y_i + Dxy_i(N_x+1:N_x+N_y);
                
                xy = np.vstack((x_i,y_i))
                max_relative = 0.0
                for it_var in range(N_x+N_y):
                    abs_value = np.abs(xy[it_var,0])
                    if abs_value < 0.001:
                        abs_value = 0.001
                                             
                    relative_error = np.abs(Dxy_i[it_var,0])/abs_value
                    
                    if relative_error > max_relative: max_relative = relative_error
                    
                if max_relative<struct[i].itol:
                    
                    break
                
                # if iter>struct[i].imax-2:
                    
                #     print('Convergence problem')

            struct[i].x[:] = x_i
            struct[i].y[:] = y_i
                
        # channels 
        if it >= it_store*decimation: 
            struct[i]['T'][it_store+1] = t 
            struct[i].X[it_store+1,:] = struct[i].x[:,0] 
            struct[i].Y[it_store+1,:] = struct[i].y[:,0]
            struct[i].iters[it_store+1,0] = iter
            struct[i].it_store += 1 
            
    struct[i].t = t
    struct[i].it_store = it_store
    return t


# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.10.3
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
import numpy as np
import ipywidgets
import matplotlib.pyplot as plt


def intplot(file,xmin=0,xmax=200,ymin=300,ymax=550,step=1.0e-6):
    Long = 200.0
    N_pos = 200
    
    fig, axes = plt.subplots(nrows=1,ncols=1, figsize=(6, 3), dpi=100)

    data = np.load(file)

    V_s = data['X'][:,0:-1:2]
    axes.plot(range(N_pos),V_s[0,:]/1e3,'r')

    line_plot = axes.plot(range(N_pos),V_s[0,:]/1e3)
    #axes.legend()
    axes.set_xlabel('Posición (m)')
    fig.tight_layout()

    axes.set_ylim((ymin,ymax))
    axes.set_xlim((xmin,xmax))

    axes.grid(True)

    fig.tight_layout()

    sld_T  = ipywidgets.FloatSlider(orientation='horizontal',description = "Time $(\mu s)$", 
                                    value=0, min=0,max= data['t'][-1]*1e6, 
                                    step=step,continuous_update=False)

    def update(change):

        t = sld_T.value   
        it = np.searchsorted(data['t'][:,0], t*1e-6)
        line_plot[0].set_data(range(N_pos),V_s[it,:]/1e3)
        fig.canvas.draw_idle()

    sld_T.observe(update, names='value')

    layout_row1 = ipywidgets.HBox([fig.canvas])
    layout_row2 = ipywidgets.HBox([sld_T])

    layout = ipywidgets.VBox([layout_row1,layout_row2])
    return layout
# -





# -*- coding: utf-8 -*-
"""
Created on Wed Dec  3 19:35:55 2014

@author: jmmauricio
"""

#import xlrd
#from xlwt import Utils
import numpy as np

def range2np(sheet, excel_range):
    if len(excel_range.split(':'))>1:  # se lee un rango
        rango = Utils.cellrange_to_rowcol_pair(excel_range)
        datos = sheet.col_values(rango[1], rango[0], rango[2])
        np_array = np.array(datos)

    if len(excel_range.split(':'))<2: # se lee una celda
        rango = Utils.cell_to_rowcol(excel_range)
        datos = sheet.cell(rango[0], rango[1]).value
        np_array = np.array(datos) 
        
    return np_array
    



class im:
    
    def __init__(self,motor_type='abb_3kW'):
        
        self.motor_type = motor_type
        self.set_motor(self.motor_type)

    def set_motor(self,motor_type):
        self.motor_type = motor_type
        
        self.library = {
            'abb_3kW':  {'P_n': 3e3,'U_n': 400.0, 'R_1': 1.86, 'X_1': 2.2, 'R_f': 1223.0, 'X_mu': 68.0, 
                         'R_2_nom': 1.59, 'X_2_nom': 1.54, 'R_2_start': 1.24, 'X_2_start': 2.2, 
                         'R_2_max': 1.24, 'X_2_max': 1.54, 'n_nom': 1445.0, 'n_max': 1055.0, 'freq_nom':50.0, 
                         'N_pp':2, 'I_nom':6.3},
            'abb_22kW': {'P_n': 22e3,'U_n': 400.0, 'R_1': 0.16, 'X_1': 0.37, 'R_f': 369.0, 'X_mu': 16.6, 
                         'R_2_nom':   0.15, 'X_2_nom':   0.76, 
                         'R_2_start': 0.24, 'X_2_start': 0.37, 
                         'R_2_max':   0.13, 'X_2_max':   0.69, 
                         'n_nom': 1463.0, 'n_max': 1344, 'freq_nom':50.0, 'N_pp':2, 'I_nom':40.7},
            'abb_22kWw': {'P_n': 22e3,'U_n': 400.0, 'R_1': 0.16, 'X_1': 0.37, 'R_f': 369.0, 'X_mu': 16.6, 
                          'R_2_nom':   0.15, 'X_2_nom':   0.76, 
                          'R_2_start': 0.15, 'X_2_start': 0.76, 
                          'R_2_max':   0.15, 'X_2_max':   0.76, 'n_nom': 1463.0, 'n_max': 1344, 'freq_nom':50.0, 
                          'N_pp':2, 'I_nom':40.7},
            'abb_90kW':  {'P_n': 90e3,'U_n': 400.0, 'R_1': 0.027, 'X_1': 0.086, 'R_f': 115.0, 'X_mu': 3.7, 
                          'R_2_nom': 0.0237, 'X_2_nom': 0.19, 'R_2_start': 0.0646, 'X_2_start': 0.086, 
                          'R_2_max': 0.0216, 'X_2_max': 0.15, 'n_nom': 1478.0, 'n_max': 1364, 'freq_nom':50.0, 
                          'N_pp':2, 'I_nom':163}
            } 
            
        self.I_nom  = self.library[self.motor_type]['I_nom']     
        self.N_pp  = self.library[self.motor_type]['N_pp']     
        self.freq_nom = self.library[self.motor_type]['freq_nom']  
        self.freq = self.freq_nom
        self.omega_1 = 2*np.pi*self.freq_nom
        self.Omega_1 = self.omega_1/self.N_pp
        self.n_1=60.0*self.freq_nom/self.N_pp       
        self.N_pp = self.library[self.motor_type]['N_pp']  
        self.P_n = self.library[self.motor_type]['P_n']
        self.R_1 = self.library[self.motor_type]['R_1']
        self.L_1 = self.library[self.motor_type]['X_1']/self.omega_1
        self.R_f = self.library[self.motor_type]['R_f']
        self.L_mu = self.library[self.motor_type]['X_mu']/self.omega_1
        self.R_2_nominal = self.library[self.motor_type]['R_2_nom']
        self.L_2_nominal = self.library[self.motor_type]['X_2_nom']/self.omega_1
        self.R_2_start = self.library[self.motor_type]['R_2_start']
        self.L_2_start = self.library[self.motor_type]['X_2_start']/self.omega_1
        self.R_2_max = self.library[self.motor_type]['R_2_max']
        self.L_2_max = self.library[self.motor_type]['X_2_max']/self.omega_1
        self.n_nom = self.library[self.motor_type]['n_nom'] 
        self.n_max = self.library[self.motor_type]['n_max'] 
        self.operating_point = 'nominal' 
        self.Omega_nom = self.n_nom*2*np.pi/60
        self.T_u_nom =  self.P_n/self.Omega_nom    
        
        
        self.model_type = 'interpolated'
        self.U_n = self.library[self.motor_type]['U_n'] 
        self.U_1 = self.U_n
#        self.R_2 = 0.12
#        self.X_2 = 1.06   # Ω         
#
#
#        self.R_2_1 =  0.3533
#        self.R_2_2 = 0.1783    
#        
#        self.X_2_1 = 0.3740
#        self.X_2_2 = 2.3220   
#        
#        
#        self.n = 1470.0
#        
#        
#        self.T_u_n = 144.0
#        self.n_n = 1463.0
#        
#        self.R_2_nominal, self.X_2_nominal = 0.12, 1.06   # Ω 
#        self.R_2_1tart, self.X_2_1tart = 0.26, 0.35   # Ω 
#        self.Omega_1 = 1500.0*(2.0*np.pi)/60.0
#        self.V_1 = 400.0/np.sqrt(3.0)
#        self.P_n = 22.0e3
#
#        # ABB 22.0 kW 1470.0c rpm
#        self.R_2_nominal, self.X_2_nominal = 0.12, 1.06   # Ω 
#        self.R_2_start, self.X_2_start = 0.26, 0.35   # Ω 
#        
#        self.V_1 = 400.0/np.sqrt(3.0)
#        self.P_n = 22.0e3
#        
#        
#        self.P_1 = 0.0
#        self.P_u = 0.0
#        self.P_2s = 0.0
#        self.P_2r = 0.0
#        self.P_2f = 0.0
#        self.c = 0.0
#        self.c_100 = 0.0
#            
#        self.model_type = 'iyme_giti'
#        self.operating_point = 'nominal'        
#        self.N = np.hstack((np.linspace(1.0,1201.0,20),
#                            np.linspace(1201.0,1400.0,20),
#                            np.linspace(1401.0,1510.0,40)))
#                            
#
#        self.N_2 = []
#        self.T_u_2 = []
#        self.update()
#        #self.update_curve()
        
    def update(self, n_array):
        model_type = self.model_type
        freq = self.freq
        omega_1 = 2*np.pi*freq
        Omega_1 = omega_1/self.N_pp
        
        
        
        n = n_array
        R_1, X_1 = self.R_1, self.L_1*omega_1   # Ω 
        R_f, X_mu = self.R_f, self.L_mu*omega_1 # Ω
        V_1 = self.U_1/np.sqrt(3.0)
            
        Omega = n*(2.0*np.pi)/60.0
        s = (Omega_1 - Omega)/(Omega_1)
        Z_1 = R_1 + 1j*X_1
        Z_m = (R_f * 1j*X_mu) / (R_f + 1j*X_mu)

        if model_type == 'single_cage':
               
            R_2, X_2 = self.R_2, self.X_2   # Ω                                                         
            Z_2 = R_2/s + 1j*X_2          
            Z_mr = Z_m*Z_2/(Z_m + Z_2)
            Z_eq = Z_1 + Z_mr          
            I_1 = V_1/Z_eq    
            S_1 = 3.0*V_1*np.conjugate(I_1)
            E = V_1 - Z_1*I_1          
            I_2 = E/Z_2          
            P_mi = 3.0*R_2*(1.0-s)/s*(np.abs(I_2)**2)
            P_u = P_mi # en el caso del modelo de ABB          
            T_u = P_u/Omega
            self.P_2s = 3*R_1*np.abs(I_1)**2
            self.P_2r = 3*R_2*np.abs(I_2)**2
            self.P_2f = 3*np.abs(E)**2/R_f

        if model_type == 'iyme_giti':
            
            if self.operating_point == 'nominal':
                R_2, X_2 = self.R_2_nominal, self.X_2_nominal   # Ω    
            if self.operating_point == 'start':
                R_2, X_2 = self.R_2_start, self.X_2_start   # Ω                 
            if self.operating_point == 'max':
                R_2, X_2 = self.R_2_max, self.X_2_max   # Ω 
            
            P_mec = (V_1**2/R_f)*2  
            R_fe = 2*R_f

            Z_m = (R_fe * 1j*X_mu) / (R_fe + 1j*X_mu)                                      
            
            if s < 0.001:
                Z_eq = Z_1 + Z_m 
                I_1 = V_1/Z_eq 
                I_2 = 0.0
                P_mi = 0.0
                E = V_1 - Z_1*I_1
                P_u = 0.0
                P_mi = 0.0 
            else:                
                Z_2 = R_2/s + 1j*X_2          
                Z_mr = Z_m*Z_2/(Z_m + Z_2)
                Z_eq = Z_1 + Z_mr 
                I_1 = V_1/Z_eq 
                E = V_1 - Z_1*I_1 
                I_2 = E/Z_2 
                P_mi = 3.0*R_2*(1.0-s)/s*(np.abs(I_2)**2)
                P_u = P_mi - P_mec 
                
            S_1 = 3.0*V_1*np.conjugate(I_1)

                   
            T_u = P_u/Omega
            
            self.P_1 = S_1.real
            self.Q_1 = S_1.imag
            self.P_j1 = 3*R_1*np.abs(I_1)**2
            self.P_j2 = 3*R_2*np.abs(I_2)**2
            self.P_fe = 3*np.abs(E)**2/R_fe
            self.P_mi = P_mi
            self.P_mec = P_mec
            self.R_fe = R_fe           
            self.P_u = P_u

            
        if model_type == 'double_cage':
        
            R_2_1, X_2_1 = self.R_2_1, self.X_2_1   # Ω    
            R_2_2, X_2_2 = self.R_2_2, self.X_2_2   # Ω  
                        
            Z_2_1 = R_2_1/s + 1j*X_2_1
            Z_2_2 = R_2_2/s + 1j*X_2_2
            
            Z_2 = Z_2_1*Z_2_2/(Z_2_1+Z_2_2)
            
            Z_mr = Z_m*Z_2/(Z_m + Z_2)
            Z_eq = Z_1 + Z_mr
            
            S_1 = 3.0*V_1*np.conjugate(I_1)
            I_1 = V_1/Z_eq
            
            E = V_1 - Z_1*I_1
            
            I_2_1 = E/Z_2_1
            I_2_2 = E/Z_2_2
            
            P_mi_1 = 3.0*R_2_1*(1.0-s)/s*(np.abs(I_2_1)**2)
            P_mi_2 = 3.0*R_2_2*(1.0-s)/s*(np.abs(I_2_2)**2)
            
            P_mi = P_mi_1 + P_mi_2 
            P_u = P_mi# en el caso del modelo de ABB
            
            T_u = P_u/Omega
            
        if model_type == 'interpolated':
            slip_n = (self.n_1 - self.n_nom)/self.n_1
            slip_max = (self.n_1 - self.n_max)/self.n_1
            slip_array = np.array([0.0, slip_n,slip_max, 1])
            R_2_array = np.array([self.R_2_nominal,self.R_2_nominal,self.R_2_max,self.R_2_start])
            X_2_array = np.array([self.L_2_nominal,self.L_2_nominal,self.L_2_max,self.L_2_start])*omega_1
            R_2 = np.interp(s,slip_array,R_2_array) # Ω   
            X_2 = np.interp(s,slip_array,X_2_array) # Ω                                                  
            Z_2 = R_2/s + 1j*X_2          
            Z_mr = Z_m*Z_2/(Z_m + Z_2)
            Z_eq = Z_1 + Z_mr          
            I_1 = V_1/Z_eq       
            E = V_1 - Z_1*I_1          
            I_2 = E/Z_2          
            P_a = 3.0*R_2/s*(np.abs(I_2)**2)
            P_j2 = 3*R_2*np.abs(I_2)**2
            P_mi = P_a - P_j2 
            P_u = P_mi # en el caso del modelo de ABB          
            T_mi = P_a/Omega_1
            T_u = T_mi # en el caso del modelo de ABB  
            S_1 = 3.0*V_1*np.conjugate(I_1)
            
            self.R_2_array = R_2
            self.X_2_array = X_2
            self.I_1 = I_1
            self.P_1 = S_1.real
            self.Q_1 = S_1.imag
            self.P_j1 = 3*R_1*np.abs(I_1)**2
            self.P_j2 = 3*R_2*np.abs(I_2)**2
            self.P_fe = 3*np.abs(E)**2/R_f
            self.P_mi = P_mi
            self.P_a = P_a
            self.P_u = P_u  
            self.P_mec = 0.0
            
        self.T_u = T_u  
        self.i_1_m  = np.abs(I_1)
        self.P_u= P_u
        self.P_1= S_1.real
        self.c = self.P_u/self.P_n
        self.c_100 = self.c*100.0
        self.s = s
        
        return T_u
    
#    def update_curve(self):
#        
#        model_type = self.model_type
#        n = self.N
#        R_1, X_1 = self.R_1, self.X_1   # Ω 
#        R_f, X_mu = self.R_f, self.X_mu # Ω
#        n_1 = self.n_1
#        Omega_1 = n_1*(2.0*np.pi)/60.0        
#        V_1 = 400.0/np.sqrt(3.0)
#            
#        Omega = n*(2.0*np.pi)/60.0
#        s = (Omega_1 - Omega)/(Omega_1)
#        Z_1 = R_1 + 1j*X_1
#        Z_m = (R_f * 1j*X_mu) / (R_f + 1j*X_mu)
#
#        if model_type == 'iyme_giti':
#            self.n_0 = n
#            self.update()
#            T_u = self.t_u
#            I_1 = self.i_1_m 
#            
#            
#            
#        if model_type == 'single_cage':
#               
#            R_2, X_2 = self.R_2, self.X_2   # Ω                                                         
#            Z_2 = R_2/s + 1j*X_2          
#            Z_mr = Z_m*Z_2/(Z_m + Z_2)
#            Z_eq = Z_1 + Z_mr          
#            I_1 = V_1/Z_eq       
#            E = V_1 - Z_1*I_1          
#            I_2 = E/Z_2          
#            P_mi = 3.0*R_2*(1.0-s)/s*(np.abs(I_2)**2)
#            P_u = P_mi # en el caso del modelo de ABB          
#            T_u = P_u/Omega
#
#        if model_type == 'double_cage':
#        
#            R_2_1, X_2_1 = self.R_2_1, self.X_2_1   # Ω    
#            R_2_2, X_2_2 = self.R_2_2, self.X_2_2   # Ω  
#                        
#            Z_2_1 = R_2_1/s + 1j*X_2_1
#            Z_2_2 = R_2_2/s + 1j*X_2_2
#            
#            Z_2 = Z_2_1*Z_2_2/(Z_2_1+Z_2_2)
#            
#            Z_mr = Z_m*Z_2/(Z_m + Z_2)
#            Z_eq = Z_1 + Z_mr
#            
#            I_1 = V_1/Z_eq
#            
#            E = V_1 - Z_1*I_1
#            
#            I_2_1 = E/Z_2_1
#            I_2_2 = E/Z_2_2
#            
#            P_mi_1 = 3.0*R_2_1*(1.0-s)/s*(np.abs(I_2_1)**2)
#            P_mi_2 = 3.0*R_2_2*(1.0-s)/s*(np.abs(I_2_2)**2)
#            
#            P_mi = P_mi_1 + P_mi_2 
#            P_u = P_mi# en el caso del modelo de ABB
#            
#            T_u = P_u/Omega
#
#        if model_type == 'interpolated':
#            slip_n = (self.n_1 - self.n_n)/self.n_1
#            slip_array = np.array([1,0.2,slip_n,0.0])
#            R_2_array = np.array([self.R_2_start,self.R_2_max,self.R_2_nominal,self.R_2_nominal])
#            X_2_array = np.array([self.X_2_start,self.X_2_max,self.X_2_nominal,self.X_2_nominal])
#            self.R_2_nominal = self.library[self.motor_type]['R_2_nom']
#            self.X_2_nominal = self.library[self.motor_type]['X_2_nom'] 
#            self.R_2_start = self.library[self.motor_type]['R_2_start']
#            self.X_2_start = self.library[self.motor_type]['X_2_start'] 
#            self.R_2_max = self.library[self.motor_type]['R_2_max']
#            self.X_2_max = self.library[self.motor_type]['X_2_max'] 
#
#            R_2, X_2 = self.R_2, self.X_2   # Ω                                                         
#            Z_2 = R_2/s + 1j*X_2          
#            Z_mr = Z_m*Z_2/(Z_m + Z_2)
#            Z_eq = Z_1 + Z_mr          
#            I_1 = V_1/Z_eq       
#            E = V_1 - Z_1*I_1          
#            I_2 = E/Z_2          
#            P_mi = 3.0*R_2*(1.0-s)/s*(np.abs(I_2)**2)
#            P_u = P_mi # en el caso del modelo de ABB          
#            T_u = P_u/Omega
#        
#        
#        self.T_u  = T_u
#        
#        reduction = rdp(np.vstack((self.N,self.T_u)).T,epsilon=0.2)    
#        
#        self.N_2 = reduction[:,0]
#        self.T_u_2 = reduction[:,1]
#        
#        self.I_1_m  = np.abs(I_1)
#        
#        return T_u
    
    def sankey(self):
        
        import numpy as np
        import matplotlib.pyplot as plt
    
        from matplotlib.sankey import Sankey
        
        P_1_pu = round(self.P_1/self.P_1,5)  
        P_j1_pu = round(self.P_j1/self.P_1,5) 
        P_fe_pu = round(self.P_fe/self.P_1,5) 
        P_j2_pu = round(self.P_j2/self.P_1,5) 
        P_mec_pu = round(self.P_mec/self.P_1,5) 
        P_u_pu = round(self.P_u/self.P_1,3)
        
        s = self.s

        
        
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, xticks=[], yticks=[] )
        sankey = Sankey(ax=ax, unit=None)
#        sankey.add(flows=[ P_1_pu], labels=['$\sf P_1$'], orientations=[0])
#        sankey.add(flows=[-P_u_pu], labels=['$\sf P_u$'], orientations=[0])
#        sankey.add(flows=[-P_j1_pu], labels=['$\sf P_{j1}$'], orientations=[-1]) 

        if s < 0.001:
            
            P_1_pu = 1.0  
            P_j1_pu = 1.0 *round(self.P_j1/self.P_1,3) 
            P_fe_pu = 1.0 *round(self.P_fe/self.P_1,3) 
            P_j2_pu = 1.0 *round(self.P_j2/self.P_1,3) 
            P_mec_pu = 1.0 *round(self.P_mec/self.P_1,3) 
            P_u_pu = 1.0 *round(self.P_u/self.P_1,3) 
     
            P_j1_pu = self.P_j1/self.P_1 
            P_fe_pu = self.P_fe/self.P_1 
            P_j2_pu = self.P_j2/self.P_1 
            P_mec_pu = self.P_mec/self.P_1 
            P_u_pu =  self.P_u/self.P_1 

            sankey.add(flows=[P_1_pu,-P_mec_pu,-P_j2_pu,-P_fe_pu,-P_j1_pu],
                   labels=['$\sf P_1$',                      
                           '$\sf P_{mec}$',
                            '$\sf P_{J2}$',  
                            '$\sf P_{Fe}$',
                            '$\sf P_{J1}$'],
                    orientations=   [0,  -1,  -1,  -1, -1],
                    pathlengths = [0.2, 0.1, 0.1, 0.1, 0.1]).finish()

        print(s)
        if s >= 0.001:
            
            P_1_pu = 1.0  
            P_j1_pu = 1.0 *round(self.P_j1/self.P_1,3) 
            P_fe_pu = 1.0 *round(self.P_fe/self.P_1,3) 
            P_j2_pu = 1.0 *round(self.P_j2/self.P_1,3) 
            P_mec_pu = 1.0 *round(self.P_mec/self.P_1,3) 
            P_u_pu = 1.0 *round(self.P_u/self.P_1,3) 
     
            P_j1_pu = self.P_j1/self.P_1 
            P_fe_pu = self.P_fe/self.P_1 
            P_j2_pu = self.P_j2/self.P_1 
            P_mec_pu = self.P_mec/self.P_1 
            P_u_pu =  self.P_u/self.P_1 
            
#        if P_u_pu > 0.0000:
            sankey.add(flows=[P_1_pu,-P_mec_pu,-P_j2_pu,-P_fe_pu,-P_j1_pu,-P_u_pu],
                   labels=['$\sf P_1$',                      
                           '$\sf P_{mec}$',
                            '$\sf P_{J2}$',  
                            '$\sf P_{Fe}$',
                            '$\sf P_{J1}$', 
                           '$\sf P_{u}$'],
                    orientations=[0, -1, -1, -1, -1, 0],
                    pathlengths = [0.2, 0.1, 0.1, 0.1, 0.1, 0.3],).finish()
#
#        if P_u_pu < 10000.00001:                     
#            sankey.add(flows=[P_1_pu,-P_mec_pu,-P_j2_pu,-P_fe_pu,-P_j1_pu],
#                   labels=['$\sf P_1$',                      
#                           '$\sf P_{mec}$',
#                            '$\sf P_{j2}$',  
#                            '$\sf P_{fe}$',
#                            '$\sf P_{j1}$'],
#                    orientations=[0, -1, -1, -1, -1],
#                    pathlengths = [0.2, 0.1, 0.1, 0.1, 0.1],).finish
                
        fig.savefig('im_sankey_{:d}.pdf'.format(int(self.n)))
                      
       
      
        
        
    
if __name__ == "__main__":

    n_array = 1463
    mi_1 = im('abb_22kWw')
    mi_1.update(n_array)
    print(mi_1.T_u)


# To update library using abb motsize and excel file:    
#    import xlrd
#    workbook = xlrd.open_workbook('abb_4poles_3_22_90.xls')
#    ws_22kw_eq = workbook.sheet_by_name(u'1_Eq')
#    
#    ws_eq = ws_22kw_eq
#    
#    mi_1.R_1 =  range2np(ws_eq, 'E20')
#    mi_1.X_1 =  range2np(ws_eq, 'G20')
#    
#    mi_1.R_f =  range2np(ws_eq, 'G21')
#    mi_1.X_mu = range2np(ws_eq, 'E21')
#        
#    mi_1.R_2_nom = range2np(ws_eq, 'G22')
#    mi_1.X_2_nom = range2np(ws_eq, 'E22')        
#    
#    mi_1.R_2_start = range2np(ws_eq, 'G23')
#    mi_1.X_2_start = range2np(ws_eq, 'E23')    
#        
#    mi_1.R_2_max = range2np(ws_eq, 'G24')
#    mi_1.X_2_max = range2np(ws_eq, 'E24')  
#    
#    mi_1.n_nom = range2np(ws_eq, 'E17') 
#
#    motors = {'abb_3kW':{'R_1':float(mi_1.R_1),'X_1':float(mi_1.X_1),
#                         'R_fe':float(mi_1.R_f),'X_mu':float(mi_1.X_mu),
#                         'R_2_nom':float(mi_1.R_2_nom),'X_2_nom':float(mi_1.X_2_nom),
#                         'R_2_start':float(mi_1.R_2_start),'X_2_start':float(mi_1.X_2_start),
#                         'R_2_max':float(mi_1.R_2_max),'X_2_max':float(mi_1.X_2_max),
#                         'n_nom':float(mi_1.n_nom)
#                         }}     
#    print(motors)

'''
    Equivalent motor Volt/phase		231 V	R1s [Ohms]	0.16	X1s [Ohms]	0.37
			                                   Xmagnetizing [Ohms]	16.6	Rfriction+iron [Ohms]	369
			                                   X2 r nom [Ohms]	0.76	R2 r nom [Ohms]	0.15
			                                   X2 start [Ohms]	0.37	R2 start [Ohms]	0.24
                                                  X2 max [Ohms]	0.69	R2 max [Ohms]	0.13
'''
    
 



import numpy as np
import sympy as sym
from pydae.svg_tools import svg
import pydae.build_cffi as db
from pydae.grid_bpu import bpu

grid = bpu('./core/trafo_shift/trafo_shift.json')

g_list = grid.dae['g'] 
h_dict = grid.dae['h_dict']
f_list = grid.dae['f']
x_list = grid.dae['x']
params_dict = grid.dae['params_dict']


sys = {'name':'trafo_shift',
       'params_dict':params_dict,
       'f_list':f_list,
       'g_list':g_list,
       'x_list':x_list,
       'y_ini_list':grid.dae['y_ini'],
       'y_run_list':grid.dae['y_run'],
       'u_run_dict':grid.dae['u_run_dict'],
       'u_ini_dict':grid.dae['u_ini_dict'],
       'h_dict':h_dict}

bldr = db.builder(sys)
bldr.build()

import numpy as np
import matplotlib.pyplot as plt
import ipywidgets
#plt.style.use('presentation.mplstyle')
colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

import trafo_shift
grid = trafo_shift.trafo_shift_class()

def set_path_style(self,object_id,new_style):
    #style="fill:#337ab7"
    for path in self.root.findall('.//{http://www.w3.org/2000/svg}path'):
        if 'id' in path.attrib:
            if path.attrib['id'] == object_id: 
                #if 'style' in rect.attrib:
                #    rect.attrib['style'].update(new_style)
                #else:
                path.attrib['style'] = new_style
                
class dashboard(trafo_shift.trafo_shift_class):
    
    def __init__(self):
        
        super().__init__()
        
        grid = trafo_shift.trafo_shift_class()
        grid.Dt = 0.01
        grid.decimation = 1
        grid.ini({'tap_B1_B2':1.0,'ang_B1_B2':np.deg2rad(0),'P_B3':-2000e6,'Q_B3':-500e6 },'./core/trafo_shift/xy_0.json')

        #grid.run(30.0,{})
        #grid.post();
        
        self.grid = grid
        
        plt.style.use('presentation.mplstyle')
        self.colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        self.widgets()
        
    def widgets(self):
        
        grid = self.grid
        colors = self.colors
        
        self.svg_file = './core/trafo_shift/trafo_shift.svg'

        #plt.ioff()
        #plt.clf()


        self.svg_fig = svg(self.svg_file)

        self.sld_P_B3 = ipywidgets.FloatSlider(orientation='horizontal',description = "P (GW)", 
                                    value=0, min=-2,max= 2,step=0.1,continuous_update=False)
        self.sld_Q_B3 = ipywidgets.FloatSlider(orientation='horizontal',description = "Q (Gvar)", 
                                    value=0, min=-2,max= 2,step=0.1,continuous_update=False)

        self.sld_tap = ipywidgets.FloatSlider(orientation='horizontal',description = "t (pu)", 
                                    value=1, min=0.95,max=1.05,step=0.01,continuous_update=False)
        self.sld_ang = ipywidgets.FloatSlider(orientation='horizontal',description = "α (º)", 
                                    value=0, min=-30,max= 30,step=1,continuous_update=False)


        self.html_grid = ipywidgets.HTML(self.svg_fig.tostring())
        
        self.update(0)

        self.sld_P_B3.observe(self.update, names='value')
        self.sld_Q_B3.observe(self.update, names='value')
        self.sld_tap.observe( self.update, names='value')
        self.sld_ang.observe( self.update, names='value')

        self.layout = ipywidgets.HBox([self.html_grid,ipywidgets.VBox([self.sld_P_B3,self.sld_Q_B3,self.sld_tap,self.sld_ang])])
        #layout_row2 = widgets.HBox([sel_vsc,,text_post])
        #layout_row3 = widgets.HBox([sld_q_vsc_R10])


        
    def update(self,change):
        
        #options=['VSC R10-S10','VSC R14-S14','VSC I02-H02','VSC C16-D16','VSC C09-D09'],
        tap = self.sld_tap.value
        ang = self.sld_ang.value

        P_B3 = self.sld_P_B3.value*1e9
        Q_B3 = self.sld_Q_B3.value*1e9
        self.grid.ini({'tap_B1_B2':tap,'ang_B1_B2':np.deg2rad(ang),'P_B3':-P_B3,'Q_B3':-Q_B3 },'xy_0.json')


        self.svg_fig.set_text(  'tap',f't ={self.grid.get_value("tap_B1_B2"):5.2f}')
        self.svg_fig.set_text('alpha',f'α ={np.rad2deg(self.grid.get_value("ang_B1_B2")):4.1f}º')

        P_B1_B3 = self.grid.get_value("p_line_B1_B3")
        Q_B1_B3 = self.grid.get_value("q_line_B1_B3")
        S_B1_B3 = (P_B1_B3**2 + Q_B1_B3**2)**0.5
        self.svg_fig.set_text('P_B1_B3',f'P = {P_B1_B3/10:0.2f} GW')
        self.svg_fig.set_text('Q_B1_B3',f'Q = {Q_B1_B3/10:0.2f} Gvar')
        self.svg_fig.set_text('S_B1_B3',f'S = {S_B1_B3/10:0.2f} GVA')

        P_B2_B3 = self.grid.get_value("p_line_B2_B3")
        Q_B2_B3 = self.grid.get_value("q_line_B2_B3")
        S_B2_B3 = (P_B2_B3**2 + Q_B2_B3**2)**0.5
        self.svg_fig.set_text('P_B2_B3',f'P = {P_B2_B3/10:0.2f} GW')
        self.svg_fig.set_text('Q_B2_B3',f'Q = {Q_B2_B3/10:0.2f} Gvar')
        self.svg_fig.set_text('S_B2_B3',f'S = {S_B2_B3/10:0.2f} GVA')

        self.svg_fig.set_text('P_B3',f'P = {-self.grid.get_value("P_B3")/1e9:4.2f} GW')
        self.svg_fig.set_text('Q_B3',f'Q = {-self.grid.get_value("Q_B3")/1e9:4.2f} Gvar')


        S_max = 1.1

        #svg_fig.set_color('path','line_B1_B3',[1,1,0])
        S = S_B1_B3/10
        red  = np.clip(255*((S - 0.0)/(S_max - 0))**4,0,255)
        width = 0.5+S/5
        set_path_style(self.svg_fig,'line_B1_B3',f'fill:none;stroke:#{int(red):02x}0000;stroke-width:{width};stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1')

        #set_path_style(svg_fig,'line_B2_B3','fill:none;stroke:#000000;stroke-width:0.529167;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1')

        #if S_B2_B3/10 > 1.02:

        S = S_B2_B3/10
        red  = np.clip(255*((S - 0.0)/(S_max - 0))**4,0,255)
        width = 0.5+S/5
        set_path_style(self.svg_fig,'line_B2_B3',f'fill:none;stroke:#{int(red):02x}0000;stroke-width:{width};stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1')

        #SVG(svg_fig.tostring())

        self.html_grid.value = self.svg_fig.tostring()

        #self.fig.canvas.draw_idle()


    def show(self):

        display(self.layout)
        
        

     